import json
from pathlib import Path
from typing import TYPE_CHECKING

import openpyxl

from pbi_core import LocalReport

from .text_elements import TextElement

if TYPE_CHECKING:
    from _typeshed import StrPath


def set_static_elements(translation_path: "StrPath", pbix_path: "StrPath") -> None:
    """We parse an excel file containing translations and create a pbix file for each language.

    The excel file must have the following structure:
    - Each worksheet represents a category (e.g., "Section", "Visual", etc.)
    - The first row contains the headers: "xpath", "field", "default", and then one column for each language
        (e.g., "en", "fr", "de").
    """
    wb = openpyxl.load_workbook(translation_path)
    languages: list[str] = [str(x) for x in next(iter(wb.worksheets[0].values))[3:]]
    processing: dict[str, list[TextElement]] = {}
    for ws in wb.worksheets:
        for row in list(ws.values)[1:]:
            for i, language in enumerate(languages):
                processing.setdefault(language, []).append(
                    TextElement(
                        category=ws.title,
                        source="layout",
                        xpath=json.loads(str(row[0])),
                        field=str(row[1]),
                        text=str(row[3 + i]),
                    ),
                )
    for language, static_elements in processing.items():
        pbix = LocalReport.load_pbix(pbix_path)
        for static_element in static_elements:
            node = pbix.static_files.layout.find_xpath(static_element.xpath)
            setattr(node, static_element.field, static_element.text)
        out_path = f"{Path(pbix_path).with_suffix('').absolute().as_posix()}_{language}.pbix"
        pbix.save_pbix(out_path)



from pydantic import ConfigDict, Field

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector

from .base import BaseVisual, ColumnProperty
from .properties.base import Expression


class CategoryAxisProperties(LayoutNode):
    class _CategoryAxisPropertiesHelper(LayoutNode):
        axisStyle: Expression | None = None
        axisType: Expression | None = None
        concatenateLabels: Expression | None = None
        end: Expression | None = None
        fontFamily: Expression | None = None
        fontSize: Expression | None = None
        gridlineShow: Expression | None = None
        innerPadding: Expression | None = None
        invertAxis: Expression | None = None
        italic: Expression | None = None
        labelColor: Expression | None = None
        labelDisplayUnits: Expression | None = None
        labelPrecision: Expression | None = None
        logAxisScale: Expression | None = None
        maxMarginFactor: Expression | None = None
        position: Expression | None = None
        preferredCategoryWidth: Expression | None = None
        show: Expression | None = None
        showAxisTitle: Expression | None = None
        start: Expression | None = None
        switchAxisPosition: Expression | None = None
        titleColor: Expression | None = None
        titleFontFamily: Expression | None = None
        titleFontSize: Expression | None = None
        titleItalic: Expression | None = None
        titleText: Expression | None = None

    properties: _CategoryAxisPropertiesHelper = Field(default_factory=_CategoryAxisPropertiesHelper)


class DataPointProperties(LayoutNode):
    class _DataPointPropertiesHelper(LayoutNode):
        borderColorMatchFill: Expression | None = None
        borderShow: Expression | None = None
        borderSize: Expression | None = None
        borderTransparency: Expression | None = None
        fill: Expression | None = None
        fillRule: Expression | None = None
        fillTransparency: Expression | None = None
        showAllDataPoints: Expression | None = None

    properties: _DataPointPropertiesHelper = Field(default_factory=_DataPointPropertiesHelper)
    selector: Selector | None = None


class GeneralProperties(LayoutNode):
    class _GeneralPropertiesHelper(LayoutNode):
        responsive: Expression | None = None

    properties: _GeneralPropertiesHelper = Field(default_factory=_GeneralPropertiesHelper)


class LabelsProperties(LayoutNode):
    class _LabelsPropertiesHelper(LayoutNode):
        backgroundColor: Expression | None = None
        backgroundTransparency: Expression | None = None
        color: Expression | None = None
        detailFontFamily: Expression | None = None
        detailItalic: Expression | None = None
        enableBackground: Expression | None = None
        enableDetailDataLabel: Expression | None = None
        enableTitleDataLabel: Expression | None = None
        fontFamily: Expression | None = None
        fontSize: Expression | None = None
        italic: Expression | None = None
        labelContainerMaxWidth: Expression | None = None
        labelContentLayout: Expression | None = None
        labelDensity: Expression | None = None
        labelDisplayUnits: Expression | None = None
        labelOverflow: Expression | None = None
        labelPosition: Expression | None = None
        labelPrecision: Expression | None = None
        optimizeLabelDisplay: Expression | None = None
        show: Expression | None = None
        showAll: Expression | None = None
        showBlankAs: Expression | None = None
        titleColor: Expression | None = None
        titleFontFamily: Expression | None = None
        titleFontSize: Expression | None = None
        titleItalic: Expression | None = None
        titleShowBlankAs: Expression | None = None
        titleTransparency: Expression | None = None
        titleUnderline: Expression | None = None
        transparency: Expression | None = None

    properties: _LabelsPropertiesHelper = Field(default_factory=_LabelsPropertiesHelper)
    selector: Selector | None = None


class LayoutProperties(LayoutNode):
    class _LayoutPropertiesHelper(LayoutNode):
        ribbonGapSize: Expression | None = None

    properties: _LayoutPropertiesHelper = Field(default_factory=_LayoutPropertiesHelper)


class LegendProperties(LayoutNode):
    class _LegendPropertiesHelper(LayoutNode):
        fontSize: Expression | None = None
        labelColor: Expression | None = None
        position: Expression | None = None
        show: Expression | None = None
        showTitle: Expression | None = None
        titleText: Expression | None = None

    properties: _LegendPropertiesHelper = Field(default_factory=_LegendPropertiesHelper)


class RibbonBandsProperties(LayoutNode):
    class _RibbonBandsPropertiesHelper(LayoutNode):
        borderColorMatchFill: Expression | None = None
        borderShow: Expression | None = None
        borderTransparency: Expression | None = None
        color: Expression | None = None
        fillTransparency: Expression | None = None
        show: Expression | None = None
        transparency: Expression | None = None

    properties: _RibbonBandsPropertiesHelper = Field(default_factory=_RibbonBandsPropertiesHelper)


class ValueAxisProperties(LayoutNode):
    class _ValueAxisPropertiesHelper(LayoutNode):
        axisScale: Expression | None = None
        bold: Expression | None = None
        end: Expression | None = None
        fontFamily: Expression | None = None
        fontSize: Expression | None = None
        gridlineColor: Expression | None = None
        gridlineShow: Expression | None = None
        invertAxis: Expression | None = None
        labelColor: Expression | None = None
        labelDisplayUnits: Expression | None = None
        labelPrecision: Expression | None = None
        logAxisScale: Expression | None = None
        show: Expression | None = None
        showAxisTitle: Expression | None = None
        start: Expression | None = None
        titleFontFamily: Expression | None = None
        titleColor: Expression | None = None
        titleFontColor: Expression | None = None
        titleFontSize: Expression | None = None
        titleText: Expression | None = None

    properties: _ValueAxisPropertiesHelper = Field(default_factory=_ValueAxisPropertiesHelper)


class XAxisReferenceLineProperties(LayoutNode):
    class _XAxisReferenceLinePropertiesHelper(LayoutNode):
        displayName: Expression | None = None
        show: Expression | None = None
        value: Expression | None = None

    properties: _XAxisReferenceLinePropertiesHelper = Field(default_factory=_XAxisReferenceLinePropertiesHelper)
    selector: Selector | None = None


class ZoomProperties(LayoutNode):
    class _ZoomPropertiesHelper(LayoutNode):
        show: Expression | None = None
        showLabels: Expression | None = None
        showTooltip: Expression | None = None

    properties: _ZoomPropertiesHelper = Field(default_factory=_ZoomPropertiesHelper)


class BarChartProperties(LayoutNode):
    categoryAxis: list[CategoryAxisProperties] = Field(default_factory=lambda: [CategoryAxisProperties()])
    dataPoint: list[DataPointProperties] = Field(default_factory=lambda: [DataPointProperties()])
    general: list[GeneralProperties] = Field(default_factory=lambda: [GeneralProperties()])
    labels: list[LabelsProperties] = Field(default_factory=lambda: [LabelsProperties()])
    layout: list[LayoutProperties] = Field(default_factory=lambda: [LayoutProperties()])
    legend: list[LegendProperties] = Field(default_factory=lambda: [LegendProperties()])
    ribbonBands: list[RibbonBandsProperties] = Field(default_factory=lambda: [RibbonBandsProperties()])
    valueAxis: list[ValueAxisProperties] = Field(default_factory=lambda: [ValueAxisProperties()])
    xAxisReferenceLine: list[XAxisReferenceLineProperties] = Field(
        default_factory=lambda: [XAxisReferenceLineProperties()],
    )
    zoom: list[ZoomProperties] = Field(default_factory=lambda: [ZoomProperties()])


class BarChart(BaseVisual):
    visualType: str = "barChart"
    model_config = ConfigDict(extra="forbid")
    columnProperties: dict[str, ColumnProperty] | None = None
    objects: BarChartProperties = Field(default_factory=BarChartProperties)



from pydantic import ConfigDict, Field

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector

from .base import BaseVisual, ColumnProperty
from .properties.base import Expression


class CategoryAxisProperties(LayoutNode):
    class _CategoryAxisPropertiesHelper(LayoutNode):
        color: Expression | None = None
        show: Expression | None = None

    properties: _CategoryAxisPropertiesHelper = Field(default_factory=_CategoryAxisPropertiesHelper)


class DataPointProperties(LayoutNode):
    class _DataPointPropertiesHelper(LayoutNode):
        fill: Expression | None = None
        showAllDataPoints: Expression | None = None

    properties: _DataPointPropertiesHelper = Field(default_factory=_DataPointPropertiesHelper)
    selector: Selector | None = None


class LabelsProperties(LayoutNode):
    class _LabelsPropertiesHelper(LayoutNode):
        color: Expression | None = None
        fontSize: Expression | None = None
        funnelLabelStyle: Expression | None = None
        labelDisplayUnits: Expression | None = None
        percentageLabelPrecision: Expression | None = None
        show: Expression | None = None

    properties: _LabelsPropertiesHelper = Field(default_factory=_LabelsPropertiesHelper)


class PercentBarLabelProperties(LayoutNode):
    class _PercentBarLabelPropertiesHelper(LayoutNode):
        color: Expression | None = None
        show: Expression | None = None

    properties: _PercentBarLabelPropertiesHelper = Field(default_factory=_PercentBarLabelPropertiesHelper)


class FunnelProperties(LayoutNode):
    categoryAxis: list[CategoryAxisProperties] = Field(default_factory=lambda: [CategoryAxisProperties()])
    dataPoint: list[DataPointProperties] = Field(default_factory=lambda: [DataPointProperties()])
    labels: list[LabelsProperties] = Field(default_factory=lambda: [LabelsProperties()])
    percentBarLabel: list[PercentBarLabelProperties] = Field(
        default_factory=lambda: [PercentBarLabelProperties()],
    )


class Funnel(BaseVisual):
    visualType: str = "funnel"
    model_config = ConfigDict(extra="forbid")
    columnProperties: dict[str, ColumnProperty] | None = None
    drillFilterOtherVisuals: bool = True
    objects: FunnelProperties = Field(default_factory=FunnelProperties)



from pydantic import ConfigDict, Field

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector

from .base import BaseVisual, ColumnProperty
from .properties.base import Expression


class BackgroundProperties(LayoutNode):
    class _BackgroundPropertiesHelper(LayoutNode):
        show: Expression | None = None
        transparency: Expression | None = None

    properties: _BackgroundPropertiesHelper = Field(default_factory=_BackgroundPropertiesHelper)


class DataPointProperties(LayoutNode):
    class _DataPointPropertiesHelper(LayoutNode):
        fill: Expression | None = None

    properties: _DataPointPropertiesHelper = Field(default_factory=_DataPointPropertiesHelper)
    selector: Selector | None = None


class GeneralProperties(LayoutNode):
    class _GeneralPropertiesHelper(LayoutNode):
        altText: Expression | None = None

    properties: _GeneralPropertiesHelper = Field(default_factory=_GeneralPropertiesHelper)


class LabelsProperties(LayoutNode):
    class _LabelsPropertiesHelper(LayoutNode):
        background: Expression | None = None
        color: Expression | None = None
        fontFamily: Expression | None = None
        fontSize: Expression | None = None
        labelDisplayUnits: Expression | None = None
        labelStyle: Expression | None = None
        overflow: Expression | None = None
        percentageLabelPrecision: Expression | None = None
        position: Expression | None = None
        show: Expression | None = None

    properties: _LabelsPropertiesHelper = Field(default_factory=_LabelsPropertiesHelper)


class LegendProperties(LayoutNode):
    class _LegendPropertiesHelper(LayoutNode):
        fontSize: Expression | None = None
        labelColor: Expression | None = None
        position: Expression | None = None
        show: Expression | None = None
        showTitle: Expression | None = None

    properties: _LegendPropertiesHelper = Field(default_factory=_LegendPropertiesHelper)


class SlicesProperties(LayoutNode):
    class _SlicesPropertiesHelper(LayoutNode):
        innerRadiusRatio: Expression | None = None

    properties: _SlicesPropertiesHelper = Field(default_factory=_SlicesPropertiesHelper)


class TitleProperties(LayoutNode):
    class _TitlePropertiesHelper(LayoutNode):
        alignment: Expression | None = None
        fontColor: Expression | None = None
        fontSize: Expression | None = None
        show: Expression | None = None
        text: Expression | None = None

    properties: _TitlePropertiesHelper = Field(default_factory=_TitlePropertiesHelper)


class DonutChartProperties(LayoutNode):
    background: list[BackgroundProperties] = Field(default_factory=lambda: [BackgroundProperties()])
    dataPoint: list[DataPointProperties] = Field(default_factory=lambda: [DataPointProperties()])
    general: list[GeneralProperties] = Field(default_factory=lambda: [GeneralProperties()])
    labels: list[LabelsProperties] = Field(default_factory=lambda: [LabelsProperties()])
    legend: list[LegendProperties] = Field(default_factory=lambda: [LegendProperties()])
    slices: list[SlicesProperties] = Field(default_factory=lambda: [SlicesProperties()])
    title: list[TitleProperties] = Field(default_factory=lambda: [TitleProperties()])


class DonutChart(BaseVisual):
    visualType: str = "donutChart"
    model_config = ConfigDict(extra="forbid")

    columnProperties: dict[str, ColumnProperty] | None = None
    drillFilterOtherVisuals: bool = True
    objects: DonutChartProperties = Field(default_factory=DonutChartProperties)



from pydantic import ConfigDict, Field

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector

from .base import BaseVisual, ColumnProperty
from .properties.base import Expression


class CategoryLabelsProperties(LayoutNode):
    class _CategoryLabelsPropertiesHelper(LayoutNode):
        color: Expression | None = None
        fontFamily: Expression | None = None
        fontSize: Expression | None = None
        show: Expression | None = None

    properties: _CategoryLabelsPropertiesHelper = Field(default_factory=_CategoryLabelsPropertiesHelper)
    selector: Selector | None = None


class GeneralProperties(LayoutNode):
    class _GeneralPropertiesHelper(LayoutNode):
        pass

    properties: _GeneralPropertiesHelper = Field(default_factory=_GeneralPropertiesHelper)


class LabelsProperties(LayoutNode):
    class _LabelsPropertiesHelper(LayoutNode):
        color: Expression | None = None
        fontFamily: Expression | None = None
        fontSize: Expression | None = None
        labelPrecision: Expression | None = None
        labelDisplayUnits: Expression | None = None
        preserveWhitespace: Expression | None = None

    properties: _LabelsPropertiesHelper = Field(default_factory=_LabelsPropertiesHelper)
    selector: Selector | None = None


class WordWrapProperties(LayoutNode):
    class _WordWrapperPropertiesHelper(LayoutNode):
        show: Expression | None = None

    properties: _WordWrapperPropertiesHelper = Field(default_factory=_WordWrapperPropertiesHelper)


class CardProperties(LayoutNode):
    categoryLabels: list[CategoryLabelsProperties] = Field(default_factory=lambda: [CategoryLabelsProperties()])
    general: list[GeneralProperties] = Field(default_factory=lambda: [GeneralProperties()])
    labels: list[LabelsProperties] = Field(default_factory=lambda: [LabelsProperties()])
    wordWrap: list[WordWrapProperties] = Field(default_factory=lambda: [WordWrapProperties()])


class Card(BaseVisual):
    visualType: str = "card"
    model_config = ConfigDict(extra="forbid")

    columnProperties: dict[str, ColumnProperty] | None = None
    drillFilterOtherVisuals: bool = True
    objects: CardProperties = Field(default_factory=CardProperties)



from pydantic import ConfigDict, Field

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector

from .base import BaseVisual
from .properties.base import Expression


class FillProperties(LayoutNode):
    class _FillPropertiesHelper(LayoutNode):
        fillColor: Expression | None = None
        image: Expression | None = None
        show: Expression | None = None
        transparency: Expression | None = None

    properties: _FillPropertiesHelper = Field(default_factory=_FillPropertiesHelper)
    selector: Selector | None = None


class IconProperties(LayoutNode):
    class _IconPropertiesHelper(LayoutNode):
        bottomMargin: Expression | None = None
        horizontalAlignment: Expression | None = None
        leftMargin: Expression | None = None
        lineColor: Expression | None = None
        lineTransparency: Expression | None = None
        lineWeight: Expression | None = None
        padding: Expression | None = None
        rightMargin: Expression | None = None
        shapeType: Expression | None = None
        show: Expression | None = None
        topMargin: Expression | None = None
        verticalAlignment: Expression | None = None

    properties: _IconPropertiesHelper = Field(default_factory=_IconPropertiesHelper)
    selector: Selector | None = None


class OutlineProperties(LayoutNode):
    class _OutlinePropertiesHelper(LayoutNode):
        lineColor: Expression | None = None
        roundEdge: Expression | None = None
        show: Expression | None = None
        transparency: Expression | None = None
        weight: Expression | None = None

    properties: _OutlinePropertiesHelper = Field(default_factory=_OutlinePropertiesHelper)
    selector: Selector | None = None


class ShapeProperties(LayoutNode):
    class _ShapePropertiesHelper(LayoutNode):
        roundEdge: Expression | None = None

    properties: _ShapePropertiesHelper = Field(default_factory=_ShapePropertiesHelper)
    selector: Selector | None = None


class TextProperties(LayoutNode):
    class _TextPropertiesHelper(LayoutNode):
        fontColor: Expression | None = None
        fontFamily: Expression | None = None
        fontSize: Expression | None = None
        horizontalAlignment: Expression | None = None
        leftMargin: Expression | None = None
        padding: Expression | None = None
        rightMargin: Expression | None = None
        show: Expression | None = None
        text: Expression | None = None
        topMargin: Expression | None = None
        verticalAlignment: Expression | None = None

    properties: _TextPropertiesHelper = Field(default_factory=_TextPropertiesHelper)
    selector: Selector | None = None


class ActionButtonProperties(LayoutNode):
    fill: list[FillProperties] = Field(default_factory=lambda: [FillProperties()])
    icon: list[IconProperties] = Field(default_factory=lambda: [IconProperties()])
    outline: list[OutlineProperties] = Field(default_factory=lambda: [OutlineProperties()])
    shape: list[ShapeProperties] = Field(default_factory=lambda: [ShapeProperties()])
    text: list[TextProperties] = Field(default_factory=lambda: [TextProperties()])


class ActionButton(BaseVisual):
    visualType: str = "actionButton"
    model_config = ConfigDict(extra="forbid")

    drillFilterOtherVisuals: bool = True
    objects: ActionButtonProperties = Field(default_factory=ActionButtonProperties)



import textwrap
from pathlib import Path
from typing import Any

import jinja2

from pbi_core.logging import get_logger
from pbi_core.pydantic import BaseValidation

logger = get_logger()

PACKAGE_DIR = Path(__file__).parents[2]
assert PACKAGE_DIR.name == "pbi_core"


class PbiCoreStartupConfig(BaseValidation):
    cert_dir: Path
    msmdsrv_ini: Path
    msmdsrv_exe: Path
    workspace_dir: Path

    def model_post_init(self, __context: Any, /) -> None:
        if not self.cert_dir.is_absolute():
            self.cert_dir = PACKAGE_DIR / self.cert_dir
        if not self.msmdsrv_ini.is_absolute():
            self.msmdsrv_ini = PACKAGE_DIR / self.msmdsrv_ini
        if not self.msmdsrv_exe.is_absolute():
            self.msmdsrv_exe = PACKAGE_DIR / self.msmdsrv_exe
        if not self.workspace_dir.is_absolute():
            self.workspace_dir = PACKAGE_DIR / self.workspace_dir

    def msmdsrv_ini_template(self) -> jinja2.Template:
        return jinja2.Template(self.msmdsrv_ini.read_text())


def get_startup_config() -> PbiCoreStartupConfig:
    config_path = PACKAGE_DIR / "local" / "settings.json"
    try:
        logger.info("Loading startup configuration", path=config_path)
        with config_path.open("r") as f:
            cfg = PbiCoreStartupConfig.model_validate_json(f.read())
            logger.info("Loaded startup configuration", path=config_path)
            return cfg
    except FileNotFoundError as e:
        logger.error("Startup configuration not found", path=config_path)
        msg = textwrap.dedent("""

        When loading a pbix file with pbi_core, the package needs one of the following:
            1. The package needs to be initialized once with "python -m pbi_core setup" to find the necessary SSAS
               files (currently broken)
            2. To be run while PowerBI Desktop is currently running, so that the SSAS server set up by PowerBI Desktop
               can be used
            3. The load_pbix function can be called with the `load_ssas=False` argument, which will not load the SSAS
               model and therefore not require the SSAS server to be set up.
        """)
        raise FileNotFoundError(msg) from e



from pydantic import ConfigDict, Field

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector

from .base import BaseVisual, ColumnProperty
from .properties.base import Expression


class BubblesProperties(LayoutNode):
    class _BubblesPropertiesHelper(LayoutNode):
        bubbleSize: Expression | None = None
        markerShape: Expression | None = None
        showSeries: Expression | None = None

    properties: _BubblesPropertiesHelper = Field(default_factory=_BubblesPropertiesHelper)
    selector: Selector | None = None


class CategoryAxisProperties(LayoutNode):
    class _CategoryAxisPropertiesHelper(LayoutNode):
        axisScale: Expression | None = None
        end: Expression | None = None
        fontFamily: Expression | None = None
        fontSize: Expression | None = None
        gridlineColor: Expression | None = None
        gridlineShow: Expression | None = None
        gridlineStyle: Expression | None = None
        innerPadding: Expression | None = None
        labelColor: Expression | None = None
        logAxisScale: Expression | None = None
        maxMarginFactor: Expression | None = None
        show: Expression | None = None
        showAxisTitle: Expression | None = None
        start: Expression | None = None
        titleColor: Expression | None = None
        titleFontFamily: Expression | None = None
        titleFontSize: Expression | None = None
        titleText: Expression | None = None
        treatNullsAsZero: Expression | None = None

    properties: _CategoryAxisPropertiesHelper = Field(default_factory=_CategoryAxisPropertiesHelper)


class CategoryLabelsProperties(LayoutNode):
    class _CategoryLabelsPropertiesHelper(LayoutNode):
        color: Expression | None = None
        enableBackground: Expression | None = None
        fontFamily: Expression | None = None
        fontSize: Expression | None = None
        show: Expression | None = None

    properties: _CategoryLabelsPropertiesHelper = Field(default_factory=_CategoryLabelsPropertiesHelper)


class ColorBorderProperties(LayoutNode):
    class _ColorBorderPropertiesHelper(LayoutNode):
        show: Expression | None = None

    properties: _ColorBorderPropertiesHelper = Field(default_factory=_ColorBorderPropertiesHelper)


class DataPointProperties(LayoutNode):
    class _DataPointPropertiesHelper(LayoutNode):
        fill: Expression | None = None
        fillRule: Expression | None = None
        legend: Expression | None = None
        showAllDataPoints: Expression | None = None
        valueAxis: Expression | None = None

    properties: _DataPointPropertiesHelper = Field(default_factory=_DataPointPropertiesHelper)
    selector: Selector | None = None


class FillPointProperties(LayoutNode):
    class _FillPointPropertiesHelper(LayoutNode):
        show: Expression | None = None
        style: Expression | None = None

    properties: _FillPointPropertiesHelper = Field(default_factory=_FillPointPropertiesHelper)


class GeneralProperties(LayoutNode):
    class _GeneralPropertiesHelper(LayoutNode):
        responsive: Expression | None = None

    properties: _GeneralPropertiesHelper = Field(default_factory=_GeneralPropertiesHelper)


class LegendProperties(LayoutNode):
    class _LegendPropertiesHelper(LayoutNode):
        fontSize: Expression | None = None
        labelColor: Expression | None = None
        position: Expression | None = None
        show: Expression | None = None
        showGradientLegend: Expression | None = None
        showTitle: Expression | None = None
        titleText: Expression | None = None

    properties: _LegendPropertiesHelper = Field(default_factory=_LegendPropertiesHelper)


class PlotAreaProperties(LayoutNode):
    class _PlotAreaPropertiesHelper(LayoutNode):
        transparency: Expression | None = None

    properties: _PlotAreaPropertiesHelper = Field(default_factory=_PlotAreaPropertiesHelper)


class ValueAxisProperties(LayoutNode):
    class _ValueAxisPropertiesHelper(LayoutNode):
        alignZeros: Expression | None = None
        axisScale: Expression | None = None
        end: Expression | None = None
        fontSize: Expression | None = None
        gridlineColor: Expression | None = None
        gridlineShow: Expression | None = None
        labelColor: Expression | None = None
        logAxisScale: Expression | None = None
        show: Expression | None = None
        showAxisTitle: Expression | None = None
        start: Expression | None = None
        switchAxisPosition: Expression | None = None
        titleColor: Expression | None = None
        titleFontFamily: Expression | None = None
        titleFontSize: Expression | None = None
        titleText: Expression | None = None
        treatNullsAsZero: Expression | None = None

    properties: _ValueAxisPropertiesHelper = Field(default_factory=_ValueAxisPropertiesHelper)


class Y1AxisReferenceLineProperties(LayoutNode):
    class _Y1AxisReferenceLinePropertiesHelper(LayoutNode):
        displayName: Expression | None = None
        lineColor: Expression | None = None
        show: Expression | None = None
        value: Expression | None = None

    properties: _Y1AxisReferenceLinePropertiesHelper = Field(default_factory=_Y1AxisReferenceLinePropertiesHelper)
    selector: Selector | None = None


class ScatterChartProperties(LayoutNode):
    bubbles: list[BubblesProperties] = Field(default_factory=lambda: [BubblesProperties()])
    categoryAxis: list[CategoryAxisProperties] = Field(default_factory=lambda: [CategoryAxisProperties()])
    categoryLabels: list[CategoryLabelsProperties] = Field(default_factory=lambda: [CategoryLabelsProperties()])
    colorBorder: list[ColorBorderProperties] = Field(default_factory=lambda: [ColorBorderProperties()])
    dataPoint: list[DataPointProperties] = Field(default_factory=lambda: [DataPointProperties()])
    fillPoint: list[FillPointProperties] = Field(default_factory=lambda: [FillPointProperties()])
    general: list[GeneralProperties] = Field(default_factory=lambda: [GeneralProperties()])
    legend: list[LegendProperties] = Field(default_factory=lambda: [LegendProperties()])
    plotArea: list[PlotAreaProperties] = Field(default_factory=lambda: [PlotAreaProperties()])
    valueAxis: list[ValueAxisProperties] = Field(default_factory=lambda: [ValueAxisProperties()])
    y1AxisReferenceLine: list[Y1AxisReferenceLineProperties] = Field(
        default_factory=lambda: [Y1AxisReferenceLineProperties()],
    )


class ScatterChart(BaseVisual):
    visualType: str = "scatterChart"
    model_config = ConfigDict(extra="forbid")

    columnProperties: dict[str, ColumnProperty] | None = None
    drillFilterOtherVisuals: bool = True
    objects: ScatterChartProperties = Field(default_factory=ScatterChartProperties)



from pydantic import ConfigDict, Field

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.expansion_state import ExpansionState
from pbi_core.static_files.layout.filters import Filter
from pbi_core.static_files.layout.selector import SelectorData

from .base import BaseVisual, ColumnProperty
from .properties.base import Expression


class SyncGroup(LayoutNode):
    groupName: str
    fieldChanges: bool
    filterChanges: bool = True


class CachedFilterDisplayItems(LayoutNode):
    id: SelectorData
    displayName: str


class DataProperties(LayoutNode):
    class _DataPropertiesHelper(LayoutNode):
        endDate: Expression | None = None
        isInvertedSelectionMode: Expression | None = None
        mode: Expression | None = None
        numericEnd: Expression | None = None
        numericStart: Expression | None = None
        startDate: Expression | None = None

    properties: _DataPropertiesHelper = Field(default_factory=_DataPropertiesHelper)


class DateProperties(LayoutNode):
    class _DatePropertiesHelper(LayoutNode):
        background: Expression | None = None
        fontColor: Expression | None = None
        fontFamily: Expression | None = None
        textSize: Expression | None = None

    properties: _DatePropertiesHelper = Field(default_factory=_DatePropertiesHelper)


class GeneralProperties(LayoutNode):
    class _GeneralPropertiesHelper(LayoutNode):
        filter: Filter | None = None
        responsive: Expression | None = None
        selfFilterEnabled: Expression | None = None
        selfFilter: Filter | None = None
        orientation: Expression | None = None
        outlineColor: Expression | None = None
        outlineWeight: Expression | None = None

    properties: _GeneralPropertiesHelper = Field(default_factory=_GeneralPropertiesHelper)


class HeaderProperties(LayoutNode):
    class _HeaderPropertiesHelper(LayoutNode):
        background: Expression | None = None
        fontColor: Expression | None = None
        fontFamily: Expression | None = None
        outlineStyle: Expression | None = None
        show: Expression | None = None
        showRestatement: Expression | None = None
        text: Expression | None = None
        textSize: Expression | None = None
        underline: Expression | None = None

    properties: _HeaderPropertiesHelper = Field(default_factory=_HeaderPropertiesHelper)


class NumericInputStyleProperties(LayoutNode):
    class _NumericInputStylePropertiesHelper(LayoutNode):
        background: Expression | None = None
        fontColor: Expression | None = None
        fontFamily: Expression | None = None
        textSize: Expression | None = None

    properties: _NumericInputStylePropertiesHelper = Field(default_factory=_NumericInputStylePropertiesHelper)


class ItemProperties(LayoutNode):
    class _ItemPropertiesHelper(LayoutNode):
        background: Expression | None = None
        bold: Expression | None = None
        expandCollapseToggleType: Expression | None = None
        fontColor: Expression | None = None
        fontFamily: Expression | None = None
        outline: Expression | None = None
        outlineColor: Expression | None = None
        outlineStyle: Expression | None = None
        padding: Expression | None = None
        steppedLayoutIndentation: Expression | None = None
        textSize: Expression | None = None

    properties: _ItemPropertiesHelper = Field(default_factory=_ItemPropertiesHelper)


class PendingChangeIconProperties(LayoutNode):
    class _PendingChangeIconPropertiesHelper(LayoutNode):
        color: Expression | None = None
        size: Expression | None = None
        tooltipLabel: Expression | None = None
        tooltipText: Expression | None = None
        transparency: Expression | None = None

    properties: _PendingChangeIconPropertiesHelper = Field(default_factory=_PendingChangeIconPropertiesHelper)


class SelectionProperties(LayoutNode):
    class _SelectionPropertiesHelper(LayoutNode):
        selectAllCheckboxEnabled: Expression | None = None
        singleSelect: Expression | None = None
        strictSingleSelect: Expression | None = None

    properties: _SelectionPropertiesHelper = Field(default_factory=_SelectionPropertiesHelper)


class SliderProperties(LayoutNode):
    class _SliderPropertiesHelper(LayoutNode):
        color: Expression | None = None
        show: Expression | None = None

    properties: _SliderPropertiesHelper = Field(default_factory=_SliderPropertiesHelper)


class SlicerProperties(LayoutNode):
    date: list[DateProperties] = Field(default_factory=lambda: [DateProperties()])
    data: list[DataProperties] = Field(default_factory=lambda: [DataProperties()])
    general: list[GeneralProperties] = Field(default_factory=lambda: [GeneralProperties()])
    header: list[HeaderProperties] = Field(default_factory=lambda: [HeaderProperties()])
    items: list[ItemProperties] = Field(default_factory=lambda: [ItemProperties()])
    numericInputStyle: list[NumericInputStyleProperties] = Field(
        default_factory=lambda: [NumericInputStyleProperties()],
    )
    pendingChangesIcon: list[PendingChangeIconProperties] = Field(
        default_factory=lambda: [PendingChangeIconProperties()],
    )
    selection: list[SelectionProperties] = Field(default_factory=lambda: [SelectionProperties()])
    slider: list[SliderProperties] = Field(default_factory=lambda: [SliderProperties()])


class Slicer(BaseVisual):
    visualType: str = "slicer"
    model_config = ConfigDict(extra="forbid")
    columnProperties: dict[str, ColumnProperty] | None = None
    syncGroup: SyncGroup | None = None
    cachedFilterDisplayItems: list[CachedFilterDisplayItems] | None = None
    expansionStates: list[ExpansionState] | None = None
    objects: SlicerProperties = Field(default_factory=SlicerProperties)



from pydantic import ConfigDict, Field

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector

from .base import BaseVisual, ColumnProperty
from .properties.base import Expression


class DataPointProperties(LayoutNode):
    class _DataPointPropertiesHelper(LayoutNode):
        fill: Expression | None = None
        showAllDataPoints: Expression | None = None

    properties: _DataPointPropertiesHelper = Field(default_factory=_DataPointPropertiesHelper)
    selector: Selector | None = None


class LabelsProperties(LayoutNode):
    class _LabelsPropertiesHelper(LayoutNode):
        color: Expression | None = None
        labelDisplayUnits: Expression | None = None
        labelPrecision: Expression | None = None
        labelStyle: Expression | None = None
        percentageLabelPrecision: Expression | None = None
        show: Expression | None = None

    properties: _LabelsPropertiesHelper = Field(default_factory=_LabelsPropertiesHelper)


class LegendProperties(LayoutNode):
    class _LegendPropertiesHelper(LayoutNode):
        position: Expression | None = None
        show: Expression | None = None

    properties: _LegendPropertiesHelper = Field(default_factory=_LegendPropertiesHelper)


class PieChartProperties(LayoutNode):
    dataPoint: list[DataPointProperties] = Field(default_factory=list[DataPointProperties()])
    labels: list[LabelsProperties] = Field(default_factory=list[LabelsProperties()])
    legend: list[LegendProperties] = Field(default_factory=list[LegendProperties()])


class PieChart(BaseVisual):
    visualType: str = "pieChart"
    model_config = ConfigDict(extra="forbid")
    columnProperties: dict[str, ColumnProperty] | None = None
    objects: PieChartProperties = Field(default_factory=PieChartProperties)



from .get_static_elements import get_static_elements
from .main import get_source_text_elements
from .set_static_elements import set_static_elements
from .text_elements import TextElement, TextElements

__all__ = [
    "TextElement",
    "TextElements",
    "get_source_text_elements",
    "get_static_elements",
    "set_static_elements",
]



import json
import zipfile
from dataclasses import dataclass
from typing import TYPE_CHECKING
from zipfile import ZipFile

import bs4

from .file_classes import Connections, DiagramLayout, Metadata, Settings, parse_diagram_layout
from .file_classes.theme import Theme
from .layout.layout import Layout

if TYPE_CHECKING:
    from _typeshed import StrPath

LAYOUT_ENCODING = "utf-16-le"


@dataclass
class Version:
    major: int
    minor: int

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}"


class StaticFiles:
    content_types: bs4.BeautifulSoup
    connections: Connections | None
    # no datamodel, that's handled by the ssas folder
    diagram_layout: DiagramLayout
    layout: Layout
    metadata: Metadata
    version: Version
    security_bindings: bytes
    settings: Settings
    themes: dict[str, Theme]

    def __init__(  # noqa: PLR0913, PLR0917
        self,
        content_types: bs4.BeautifulSoup,
        connections: Connections | None,
        diagram_layout: DiagramLayout,
        layout: Layout,
        metadata: Metadata,
        version: Version,
        security_bindings: bytes,
        settings: Settings,
        themes: dict[str, Theme],
    ) -> None:
        self.content_types = content_types
        self.connections = connections
        self.diagram_layout = diagram_layout
        self.layout = layout
        self.metadata = metadata
        self.version = version
        self.security_bindings = security_bindings
        self.settings = settings
        self.themes = themes

    def __repr__(self) -> str:
        return f"StaticFiles(version={self.version})"

    @staticmethod
    def load_pbix(path: "StrPath") -> "StaticFiles":  # noqa: PLR0914
        """Load static report information from a PBIX file."""
        with ZipFile(path, mode="r") as zip_file:
            themes: dict[str, Theme] = {}
            theme_paths = [
                x.split("/")[-1]
                for x in zip_file.namelist()
                if x.startswith("Report/StaticResources/SharedResources/BaseThemes")
            ]
            for theme_path in theme_paths:
                theme_json = json.loads(
                    zip_file.read(
                        f"Report/StaticResources/SharedResources/BaseThemes/{theme_path}",
                    ).decode("utf-8"),
                )
                themes[theme_path] = Theme.model_validate(theme_json)

            layout_json = json.loads(zip_file.read("Report/Layout").decode(LAYOUT_ENCODING))
            layout = Layout.model_validate(layout_json)

            if "Connections" in zip_file.namelist():
                connections_json = json.loads(zip_file.read("Connections").decode("utf-8"))
                connections = Connections.model_validate(connections_json)
            else:
                connections = None

            major, minor = zip_file.read("Version").decode(LAYOUT_ENCODING).split(".")
            version = Version(int(major), int(minor))
            try:
                security_bindings = zip_file.read("SecurityBindings")
            except KeyError:  # This can occur when the PBIX was generated by pbi_core since the file is optional
                security_bindings = b""

            content_types = bs4.BeautifulSoup(
                zip_file.read("[Content_Types].xml").decode("utf-8"),
                "xml",
            )

            metadata_json = json.loads(zip_file.read("Metadata").decode(LAYOUT_ENCODING))
            metadata = Metadata.model_validate(metadata_json)

            diagram_json = json.loads(zip_file.read("DiagramLayout").decode(LAYOUT_ENCODING))
            diagram_layout = parse_diagram_layout(diagram_json)

            settings_json = json.loads(zip_file.read("Settings").decode(LAYOUT_ENCODING))
            settings = Settings.model_validate(settings_json)

        return StaticFiles(
            content_types,
            connections,
            diagram_layout,
            layout,
            metadata,
            version,
            security_bindings,
            settings,
            themes,
        )

    def save_pbix(self, path: "StrPath") -> None:
        """Saves report to file.

        We use "a" as the write mode because the ssas.save_pbix method already
        creates a PBIX with the datamodel subfile.
        """
        data: dict[str, bytes] = {}
        with zipfile.ZipFile(path, "r", compression=zipfile.ZIP_DEFLATED) as zf:
            for f in zf.namelist():
                data[f] = zf.read(f)

        with zipfile.ZipFile(path, "w", compression=zipfile.ZIP_DEFLATED) as zf:
            if self.connections is not None:
                zf.writestr("Connections", self.connections.model_dump_json().encode("utf-8"))
            zf.writestr("Version", str(self.version).encode(LAYOUT_ENCODING))
            for name, layout in [
                ("DiagramLayout", self.diagram_layout),
                ("Report/Layout", self.layout),
                ("Settings", self.settings),
                ("Metadata", self.metadata),
            ]:
                zf.writestr(
                    name,
                    layout.model_dump_json().encode(LAYOUT_ENCODING),
                )

            for path_name, theme_info in self.themes.items():
                zf.writestr(
                    f"Report/StaticResources/SharedResources/BaseThemes/{path_name}",
                    theme_info.model_dump_json().encode("utf-8"),
                )
            zf.writestr("DataModel", data["DataModel"])
            zf.writestr("[Content_Types].xml", str(self.content_types).encode("utf8"))



from .logging import get_logger
from .ssas.model_tables.column import Column
from .ssas.model_tables.measure import Measure
from .static_files.model_references import ModelColumnReference, ModelMeasureReference

logger = get_logger()


def column_finder(c: Column, reference: ModelColumnReference) -> bool:
    match = c.explicit_name == reference.column and c.table().name == reference.table
    logger.debug(
        "column_finder",
        column=c.explicit_name,
        table=c.table().name,
        reference_table=reference.table,
        reference_column=reference.column,
        match=match,
    )
    return match


def measure_finder(m: Measure, reference: ModelMeasureReference) -> bool:
    match = m.name == reference.measure and m.table().name == reference.table
    logger.debug(
        "measure_finder",
        measure=m.name,
        table=m.table().name,
        reference_table=reference.table,
        reference_measure=reference.measure,
        match=match,
    )
    return match



from pydantic import ConfigDict, Field

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector

from .base import BaseVisual, ColumnProperty
from .properties.base import Expression


class CategoryAxisProperties(LayoutNode):
    class _CategoryAxisPropertiesHelper(LayoutNode):
        axisType: Expression | None = None

    properties: _CategoryAxisPropertiesHelper = Field(default_factory=_CategoryAxisPropertiesHelper)


class DataPointProperties(LayoutNode):
    class _DataPointPropertiesHelper(LayoutNode):
        fill: Expression | None = None
        fillRule: Expression | None = None
        showAllDataPoints: Expression | None = None

    properties: _DataPointPropertiesHelper = Field(default_factory=_DataPointPropertiesHelper)
    selector: Selector | None = None


class LabelsProperties(LayoutNode):
    class _LabelsPropertiesHelper(LayoutNode):
        backgroundColor: Expression | None = None
        backgroundTransparency: Expression | None = None
        color: Expression | None = None
        enableBackground: Expression | None = None
        fontSize: Expression | None = None
        labelDisplayUnits: Expression | None = None
        labelOrientation: Expression | None = None
        labelPosition: Expression | None = None
        show: Expression | None = None
        showAll: Expression | None = None

    properties: _LabelsPropertiesHelper = Field(default_factory=_LabelsPropertiesHelper)
    selector: Selector | None = None


class LegendProperties(LayoutNode):
    class _LegendPropertiesHelper(LayoutNode):
        legendMarkerRendering: Expression | None = None
        position: Expression | None = None
        show: Expression | None = None

    properties: _LegendPropertiesHelper = Field(default_factory=_LegendPropertiesHelper)
    selector: Selector | None = None


class LineStylesProperties(LayoutNode):
    class _LineStylesPropertiesHelper(LayoutNode):
        lineStyle: Expression | None = None
        markerShape: Expression | None = None
        shadeArea: Expression | None = None
        showMarker: Expression | None = None
        showSeries: Expression | None = None
        stepped: Expression | None = None
        strokeWidth: Expression | None = None

    properties: _LineStylesPropertiesHelper = Field(default_factory=_LineStylesPropertiesHelper)
    selector: Selector | None = None


class SmallMultiplesLayoutProperties(LayoutNode):
    class _SmallMultiplesLayoutPropertiesHelper(LayoutNode):
        gridLineColor: Expression | None = None
        gridLineStyle: Expression | None = None
        gridLineType: Expression | None = None
        gridPadding: Expression | None = None
        rowCount: Expression | None = None

    properties: _SmallMultiplesLayoutPropertiesHelper = Field(default_factory=_SmallMultiplesLayoutPropertiesHelper)


class SubheaderProperties(LayoutNode):
    class _SubheaderPropertiesHelper(LayoutNode):
        fontSize: Expression | None = None

    properties: _SubheaderPropertiesHelper = Field(default_factory=_SubheaderPropertiesHelper)


class ValueAxisProperties(LayoutNode):
    class _ValueAxisPropertiesHelper(LayoutNode):
        alignZeros: Expression | None = None
        end: Expression | None = None
        gridlineShow: Expression | None = None
        secEnd: Expression | None = None
        secShow: Expression | None = None
        secStart: Expression | None = None
        start: Expression | None = None
        show: Expression | None = None

    properties: _ValueAxisPropertiesHelper = Field(default_factory=_ValueAxisPropertiesHelper)
    selector: Selector | None = None


class LineStackedColumnComboChartProperties(LayoutNode):
    categoryAxis: list[CategoryAxisProperties] = Field(default_factory=lambda: [CategoryAxisProperties()])
    dataPoint: list[DataPointProperties] = Field(default_factory=lambda: [DataPointProperties()])
    labels: list[LabelsProperties] = Field(default_factory=lambda: [LabelsProperties()])
    legend: list[LegendProperties] = Field(default_factory=lambda: [LegendProperties()])
    lineStyles: list[LineStylesProperties] = Field(default_factory=lambda: [LineStylesProperties()])
    smallMultiplesLayout: list[SmallMultiplesLayoutProperties] = Field(
        default_factory=lambda: [SmallMultiplesLayoutProperties()],
    )
    subheader: list[SubheaderProperties] = Field(default_factory=lambda: [SubheaderProperties()])
    valueAxis: list[ValueAxisProperties] = Field(default_factory=lambda: [ValueAxisProperties()])


class LineStackedColumnComboChart(BaseVisual):
    visualType: str = "lineStackedColumnComboChart"
    model_config = ConfigDict(extra="forbid")

    columnProperties: dict[str, ColumnProperty] | None = None
    drillFilterOtherVisuals: bool = True
    objects: LineStackedColumnComboChartProperties = Field(default_factory=LineStackedColumnComboChartProperties)



from pydantic import ConfigDict, Field

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector

from .base import BaseVisual, ColumnProperty
from .properties.base import Expression


class ColumnFormattingProperties(LayoutNode):
    class _ColumnFormattingPropertiesHelper(LayoutNode):
        class _DataBarsProperties(LayoutNode):
            axisColor: Expression | None = None
            hideText: Expression | None = None
            negativeColor: Expression | None = None
            positiveColor: Expression | None = None
            reverseDirection: Expression | None = None

        alignment: Expression | None = None
        backColor: Expression | None = None
        dataBars: _DataBarsProperties | None = None
        fontColor: Expression | None = None
        labelDisplayUnits: Expression | None = None
        labelPrecision: Expression | None = None
        styleHeader: Expression | None = None
        styleValues: Expression | None = None

    properties: _ColumnFormattingPropertiesHelper = Field(default_factory=_ColumnFormattingPropertiesHelper)
    selector: Selector | None = None


class ColumnHeadersProperties(LayoutNode):
    class _ColumnHeadersPropertiesHelper(LayoutNode):
        alignment: Expression | None = None
        autoSizeColumnWidth: Expression | None = None
        backColor: Expression | None = None
        bold: Expression | None = None
        fontColor: Expression | None = None
        fontFamily: Expression | None = None
        fontSize: Expression | None = None
        outline: Expression | None = None
        outlineStyle: Expression | None = None
        underline: Expression | None = None
        wordWrap: Expression | None = None

    properties: _ColumnHeadersPropertiesHelper = Field(default_factory=_ColumnHeadersPropertiesHelper)
    selector: Selector | None = None


class ColumnWidthProperties(LayoutNode):
    class _ColumnWidthPropertiesHelper(LayoutNode):
        value: Expression | None = None

    properties: _ColumnWidthPropertiesHelper = Field(default_factory=_ColumnWidthPropertiesHelper)
    selector: Selector | None = None


class GeneralProperties(LayoutNode):
    class _GeneralPropertiesHelper(LayoutNode):
        pass

    properties: _GeneralPropertiesHelper = Field(default_factory=_GeneralPropertiesHelper)


class GridProperties(LayoutNode):
    class _GridPropertiesHelper(LayoutNode):
        gridHorizontal: Expression | None = None
        gridHorizontalColor: Expression | None = None
        gridHorizontalWeight: Expression | None = None
        gridVertical: Expression | None = None
        gridVerticalColor: Expression | None = None
        gridVerticalWeight: Expression | None = None
        imageHeight: Expression | None = None
        outlineColor: Expression | None = None
        outlineWeight: Expression | None = None
        rowPadding: Expression | None = None
        textSize: Expression | None = None

    properties: _GridPropertiesHelper = Field(default_factory=_GridPropertiesHelper)
    selector: Selector | None = None


class TotalProperties(LayoutNode):
    class _TotalPropertiesHelper(LayoutNode):
        fontColor: Expression | None = None
        fontFamily: Expression | None = None
        fontSize: Expression | None = None
        outline: Expression | None = None
        totals: Expression | None = None

    properties: _TotalPropertiesHelper = Field(default_factory=_TotalPropertiesHelper)
    selector: Selector | None = None


class ValuesProperties(LayoutNode):
    class _ValuesPropertiesHelper(LayoutNode):
        backColor: Expression | None = None
        backColorPrimary: Expression | None = None
        backColorSecondary: Expression | None = None
        fontColor: Expression | None = None
        fontColorPrimary: Expression | None = None
        fontFamily: Expression | None = None
        fontSize: Expression | None = None
        outline: Expression | None = None
        underline: Expression | None = None
        urlIcon: Expression | None = None
        wordWrap: Expression | None = None

    properties: _ValuesPropertiesHelper = Field(default_factory=_ValuesPropertiesHelper)
    selector: Selector | None = None


class TableChartColumnProperties(LayoutNode):
    columnFormatting: list[ColumnFormattingProperties] = Field(
        default_factory=lambda: [ColumnFormattingProperties()],
    )
    columnHeaders: list[ColumnHeadersProperties] = Field(default_factory=lambda: [ColumnHeadersProperties()])
    columnWidth: list[ColumnWidthProperties] = Field(default_factory=lambda: [ColumnWidthProperties()])
    general: list[GeneralProperties] = Field(default_factory=lambda: [GeneralProperties()])
    grid: list[GridProperties] = Field(default_factory=lambda: [GridProperties()])
    total: list[TotalProperties] = Field(default_factory=lambda: [TotalProperties()])
    values: list[ValuesProperties] = Field(default_factory=lambda: [ValuesProperties()])


class TableChart(BaseVisual):
    visualType: str = "tableEx"
    objects: TableChartColumnProperties = Field(default_factory=TableChartColumnProperties)
    model_config = ConfigDict(extra="forbid")
    columnProperties: dict[str, ColumnProperty] | None = None



from typing import TYPE_CHECKING

from pbi_core.logging import get_logger
from pbi_core.report.base import BaseReport
from pbi_core.ssas.server import LocalTabularModel, get_or_create_local_server

logger = get_logger()

if TYPE_CHECKING:
    from _typeshed import StrPath


class LocalSsasReport(BaseReport):
    """An instance of the SSAS-based components in a PowerBI report from a local PBIX file.

    Examples:
        ```python
        from pbi_core import LocalReport

        report = LocalReport.load_pbix("example.pbix")
        report.save_pbix("example_out.pbix")
        ```

    """

    _source_path: "StrPath"
    """Since this class doesn't load the full PBIX, we need to keep track of the source path for saving later"""
    ssas: LocalTabularModel
    """An instance of a local SSAS Server"""

    def __init__(self, ssas: LocalTabularModel, source_path: "StrPath") -> None:
        self.ssas = ssas
        self._source_path = source_path

    @staticmethod
    def load_pbix(path: "StrPath", *, kill_ssas_on_exit: bool = True) -> "LocalSsasReport":
        """Creates a ``LocalReport`` instance from a PBIX file.

        Args:
                path (StrPath): The absolute or local path to the PBIX report
                kill_ssas_on_exit (bool, optional): The LocalReport object depends on a ``msmdsrv.exe`` process that is
                    independent of the Python session process. If this function creates a new ``msmdsrv.exe`` instance
                    and kill_ssas_on_exit is true, the process will be killed on exit.

        Examples:
            ```python

               from pbi_core import LocalReport

               report = LocalReport.load_pbix("example.pbix")
            ```

        Returns:
                LocalReport: the local PBIX class

        """
        logger.info("Loading PBIX SSAS", path=path)
        server = get_or_create_local_server(kill_on_exit=kill_ssas_on_exit)
        ssas = server.load_pbix(path)
        return LocalSsasReport(ssas=ssas, source_path=path)

    def save_pbix(self, path: "StrPath", *, sync_ssas_changes: bool = True) -> None:
        """Creates a new PBIX with the information in this class to the given path.

        Examples:
            ```python

               from pbi_core import LocalReport

               report = LocalReport.load_pbix("example.pbix")
            ```
               report.save_pbix("example_out.pbix")

        Args:
            path (StrPath): the path (relative or absolute) to save the PBIX to
            sync_ssas_changes (bool, optional): whether to sync changes made in the SSAS model back to the PBIX file

        """
        if sync_ssas_changes:
            self.ssas.sync_to()
        self.ssas.save_pbix(path)



from collections.abc import Callable
from enum import Enum
from typing import TYPE_CHECKING, Any, Literal, TypeVar

from pbi_core.lineage import LineageNode
from pbi_core.pydantic.main import BaseValidation

if TYPE_CHECKING:
    from pbi_core.ssas.server import BaseTabularModel

LAYOUT_ENCODING = "utf-16-le"


T = TypeVar("T", bound="LayoutNode")


class LayoutNode(BaseValidation):
    _name_field: str | None = None  # name of the field used to populate __repr__

    @staticmethod
    def serialize_helper(value: Any) -> Any:
        """Helper function to serialize a value.

        We need to separate from the main function to handle cases where there is a list of
        dictionaries such as the visual container properties.
        """
        if hasattr(value, "serialize"):
            return value.serialize()
        if isinstance(value, list):
            return [LayoutNode.serialize_helper(val) for val in value]
        if isinstance(value, dict):
            return {key: LayoutNode.serialize_helper(val) for key, val in value.items()}
        if isinstance(value, Enum):
            return value.name
        return value

    def serialize(self) -> dict[str, Any]:
        """Serialize the node to a dictionary.

        Differs from the model_dump_json method in that it does not convert the JSON models back to strings.
        """
        ret = {}
        for field in self.model_fields:
            ret[field] = self.serialize_helper(getattr(self, field))
        return ret

    def pbi_core_name(self) -> str:
        raise NotImplementedError

    def find_all(
        self,
        cls_type: type[T],
        attributes: dict[str, Any] | Callable[[T], bool] | None = None,
    ) -> list["T"]:
        ret: list[T] = []
        if attributes is None:
            attribute_lambda: Callable[[T], bool] = lambda _: True  # noqa: E731
        elif isinstance(attributes, dict):
            attribute_lambda = lambda x: all(  # noqa: E731
                getattr(x, field_name) == field_value for field_name, field_value in attributes.items()
            )
        else:
            attribute_lambda = attributes
        if isinstance(self, cls_type) and attribute_lambda(self):
            ret.append(self)
        for child in self._children():
            ret.extend(child.find_all(cls_type, attributes))
        return ret

    def find(self, cls_type: type[T], attributes: dict[str, Any] | Callable[[T], bool] | None = None) -> "T":
        if attributes is None:
            attribute_lambda: Callable[[T], bool] = lambda _: True  # noqa: E731
        elif isinstance(attributes, dict):
            attribute_lambda = lambda x: all(  # noqa: E731
                getattr(x, field_name) == field_value for field_name, field_value in attributes.items()
            )
        else:
            attribute_lambda = attributes
        if isinstance(self, cls_type) and attribute_lambda(self):
            return self
        for child in self._children():
            try:
                return child.find(cls_type, attributes)
            except ValueError:
                pass
        msg = f"Object not found: {cls_type}"
        raise ValueError(msg)

    def _children(self) -> list["LayoutNode"]:
        ret: list[LayoutNode] = []
        for attr in dir(self):
            if attr.startswith("_"):
                continue
            child_candidate: list[Any] | dict[str, Any] | LayoutNode | int | str = getattr(self, attr)
            if isinstance(child_candidate, list):
                ret.extend(val for val in child_candidate if isinstance(val, LayoutNode))
            elif isinstance(child_candidate, dict):
                ret.extend(val for val in child_candidate.values() if isinstance(val, LayoutNode))
            elif isinstance(child_candidate, LayoutNode):
                ret.append(child_candidate)
        return ret

    def __str__(self) -> str:
        if self._name_field is not None:
            name = getattr(self, self._name_field)
            if callable(name):
                name = name()
            return f"{self.__class__.__name__}({name})"
        return super().__str__()

    def get_lineage(
        self,
        lineage_type: Literal["children", "parents"],
        tabular_model: "BaseTabularModel",
    ) -> LineageNode:
        raise NotImplementedError

    def find_xpath(self, xpath: list[str | int]) -> "LayoutNode":
        """Find a node in the layout using an XPath-like list of attributes.

        Note: This method currently uses a DFS approach to find the node.
            Eventually, I'll find a way to type-safely include element parents in the LayoutNode.

        Raises:
            TypeError: If the XPath is invalid or if the node is not found.

        """
        if len(xpath) == 0:
            return self

        next_step = xpath.pop(0)
        if isinstance(next_step, int):
            msg = f"Cannot index {self.__class__.__name__} with an integer: {next_step}"
            raise TypeError(msg)
        attr = getattr(self, next_step)

        while isinstance(attr, (dict, list)):
            next_step = xpath.pop(0)
            attr = attr[next_step]  # pyright: ignore[reportCallIssue, reportArgumentType]

        if not isinstance(attr, LayoutNode):
            msg = f"Cannot index {self.__class__.__name__} with a non-LayoutNode: {attr}"
            raise TypeError(msg)
        return attr.find_xpath(xpath)

    def get_xpath(self, parent: "LayoutNode") -> list[str | int]:
        """Get the [XPath](https://developer.mozilla.org/en-US/docs/Web/XML/XPath) of this node.

        Args:
            parent (LayoutNode): The parent node to which the XPath is relative.

        Raises:
            ValueError: If the node is not found in the parent.

        """
        ret = _get_xpath(parent, self)
        if ret is None:
            msg = f"Node {self.pbi_core_name()} not found in parent {parent.pbi_core_name()}"
            raise ValueError(msg)
        return ret

    def pprint(self, indent: int = 4) -> None:
        ret = self.model_dump_json(indent=indent)
        ret = ret.replace('"', "").replace(":", "=").replace("{", "(").replace("}", ")")
        ret = self.__class__.__name__ + ret
        print(ret)


def _get_xpath(  # noqa: C901  # too complex, but it's actually not that complex
    parent: LayoutNode | list | dict,
    child: LayoutNode,
    xpath: list[str | int] | None = None,
) -> list[str | int] | None:
    def _xpath_pydantic(parent: LayoutNode, child: LayoutNode, xpath: list[str | int]) -> list[str | int] | None:
        for attr in parent.__pydantic_fields__:
            val = getattr(parent, attr)
            ret = _get_xpath(val, child, xpath=[*xpath, attr])
            if ret is not None:
                return ret
        return None

    def _xpath_list(parent: list, child: LayoutNode, xpath: list[str | int]) -> list[str | int] | None:
        for i, val in enumerate(parent):
            ret = _get_xpath(val, child, xpath=[*xpath, i])
            if ret is not None:
                return ret
        return None

    def _xpath_dict(parent: dict, child: LayoutNode, xpath: list[str | int]) -> list[str | int] | None:
        for key, val in parent.items():
            ret = _get_xpath(val, child, xpath=[*xpath, key])
            if ret is not None:
                return ret
        return None

    xpath = xpath or []
    # print(id(parent), id(child), child.pbi_core_name(), xpath)
    if parent is child:
        return xpath

    if isinstance(parent, LayoutNode):
        return _xpath_pydantic(parent, child, xpath)
    if isinstance(parent, list):
        return _xpath_list(parent, child, xpath)
    if isinstance(parent, dict):
        return _xpath_dict(parent, child, xpath)
    return None



import json
from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal

import openpyxl

from pbi_core.logging import get_logger
from pbi_core.report.local.main import LocalReport
from pbi_core.static_files.layout.sources.literal import LiteralSource, serialize_literal

if TYPE_CHECKING:
    from _typeshed import StrPath
logger = get_logger()


@dataclass
class TextElement:
    category: str
    """Used to group elements into different Excel sheets"""
    source: Literal["layout", "ssas"]
    """Used to identify whether the xpath should apply to the frontend or backend"""
    xpath: list[str | int]
    """Used to identify the entity to update"""
    field: str
    """Used to identify the field of the updating entity"""
    text: str


class TextElements:
    language: str = "source_text"
    text_elements: list[TextElement]

    def __init__(self, text_elements: list[TextElement] | None = None, language: str = "source_text") -> None:
        self.text_elements = text_elements or []
        self.language = language

    def to_csv(self) -> None:
        pass

    def _grouped(self) -> dict[str, list[TextElement]]:
        grouped: dict[str, list[TextElement]] = {}
        for element in self.text_elements:
            grouped.setdefault(element.category, []).append(element)
        return grouped

    def set_elements(self, report: LocalReport) -> None:
        """Updates the text elements."""
        for text_element in self.text_elements:
            if text_element.source == "layout":
                node = report.static_files.layout.find_xpath(text_element.xpath)
                if isinstance(node, LiteralSource):
                    node.Literal.Value = serialize_literal(text_element.text)
                else:
                    setattr(node, text_element.field, text_element.text)
            elif text_element.source == "ssas":
                group, idx = text_element.xpath
                assert isinstance(group, str)
                assert isinstance(idx, int)
                group = getattr(report.ssas, group)
                entity = group.find(idx)
                setattr(entity, text_element.field, text_element.text)
        if any(te.source == "ssas" for te in self.text_elements):
            report.ssas.sync_to()

    def to_excel(self, path: "StrPath") -> None:
        wb = openpyxl.Workbook()
        for category, objects in self._grouped().items():
            ws = wb.create_sheet(category)
            for j, name in enumerate(["source", "xpath", "field", "source_text"]):
                ws.cell(1, j + 1).value = name
            for i, obj in enumerate(objects):
                ws.cell(2 + i, 1).value = obj.source
                ws.cell(2 + i, 2).value = json.dumps(obj.xpath)
                ws.cell(2 + i, 3).value = obj.field
                ws.cell(2 + i, 4).value = obj.text
        wb.remove(wb["Sheet"])
        wb.save(path)

    @classmethod
    def from_excel(cls, path: "StrPath") -> "list[TextElements]":
        logger.info("Parsing Excel", path=path)
        wb = openpyxl.load_workbook(path)

        headers = next(wb.worksheets[0].rows)
        languages: list[str] = [str(c.value) for c in headers][4:]

        static_elements: dict[str, list[TextElement]] = {}
        for ws in wb.worksheets:
            for row in list(ws.values)[1:]:
                assert isinstance(row[0], str)
                assert row[0] in ("layout", "ssas")  # noqa: PLR6201
                # For now, mypy is not able to infer the type of row[0] when using set notation

                assert isinstance(row[1], str)
                assert isinstance(row[2], str)
                assert isinstance(row[3], str)
                xpath = json.loads(row[1])
                field = row[2]
                for i, lang in enumerate(languages):
                    text = row[4 + i]
                    assert isinstance(text, str)
                    static_elements.setdefault(lang, []).append(
                        TextElement(
                            category=ws.title,
                            source=row[0],
                            xpath=xpath,
                            field=field,
                            text=text,
                        ),
                    )
        return [
            TextElements(
                text_elements=elements,
                language=lang,
            )
            for lang, elements in static_elements.items()
        ]



from typing import Any, Literal, Protocol

from pbi_mermaid import Link, MermaidDiagram, Node
from pbi_mermaid.node import NodeShape

"""
#ffffcc, #ffcc99, #ffcccc,
#ff99cc, #ffccff, #cc99ff,
#ccccff, #99ccff, #ccffff,
#99ffcc, #ccffcc, #ccff99,
"""

CLASS_STYLES = {
    # SSAS Stuff
    "AttributeHierarchy": {"fill": "#ffffcc", "stroke": "#333", "stroke-width": "1px"},
    "Column": {"fill": "#ffcc99", "stroke": "#333", "stroke-width": "1px"},
    "Culture": {"fill": "#ffcccc", "stroke": "#333", "stroke-width": "1px"},
    "LinguisticMetadata": {"fill": "#ff99cc", "stroke": "#333", "stroke-width": "1px"},
    "Measure": {"fill": "#ffccff", "stroke": "#333", "stroke-width": "1px"},
    "Table": {"fill": "#cc99ff", "stroke": "#333", "stroke-width": "1px"},
    "Partition": {"fill": "#ccccff", "stroke": "#333", "stroke-width": "1px"},
    "Model": {"fill": "#99ccff", "stroke": "#333", "stroke-width": "1px"},
    # Visuals
    "BarChart": {"fill": "#ccffff", "stroke": "#333", "stroke-width": "1px"},
    "VisualContainer": {"fill": "#99ffcc", "stroke": "#333", "stroke-width": "1px"},
    "Section": {"fill": "#ccffcc", "stroke": "#333", "stroke-width": "1px"},
    "Layout": {"fill": "#ccff99", "stroke": "#333", "stroke-width": "1px"},
}


class LineageProtocol(Protocol):
    id: str

    def pbi_core_name(self) -> str:
        """Returns the name displayed in the PBIX report."""
        raise NotImplementedError


class LineageNode:
    """Class used to track DAX dependencies from visuals to measures to PowerQueries."""

    value: Any
    relatives: list["LineageNode"]
    lineage_type: Literal["children", "parents"]

    def __init__(
        self,
        value: Any,
        lineage_type: Literal["children", "parents"],
        relatives: list["LineageNode"] | None = None,
    ) -> None:
        """Initialize."""
        self.value = value
        self.relatives = relatives or []
        self.lineage_type = lineage_type

    @staticmethod
    def _create_node(value: LineageProtocol) -> Node:
        return Node(
            id=f"{value.__class__.__name__}-{value.id}",
            content=value.pbi_core_name(),
            style=CLASS_STYLES[value.__class__.__name__],
            shape=NodeShape.round_edge,
        )

    def _to_mermaid_helper(self, node: Node) -> tuple[list[Node], list[Link]]:
        nodes: list[Node] = [node]
        links: list[Link] = []
        for relative in self.relatives:
            child_node = self._create_node(relative.value)
            child_nodes, child_links = relative._to_mermaid_helper(child_node)
            links.append(Link(nodes[0], child_node))
            links.extend(child_links)
            nodes.extend(child_nodes)
        return nodes, links

    def to_mermaid(self) -> MermaidDiagram:
        base_node = self._create_node(self.value)
        nodes, links = self._to_mermaid_helper(base_node)
        return MermaidDiagram(title="Lineage Chart", nodes=nodes, links=links)



from pydantic import ConfigDict, Field

from pbi_core.static_files.layout._base_node import LayoutNode

from .base import BaseVisual
from .properties.base import Expression


class FillProperties(LayoutNode):
    class _FillPropertiesHelper(LayoutNode):
        fillColor: Expression | None = None
        show: Expression | None = None
        transparency: Expression | None = None

    properties: _FillPropertiesHelper = Field(default_factory=_FillPropertiesHelper)


class GeneralProperties(LayoutNode):
    class _GeneralPropertiesHelper(LayoutNode):
        shapeType: Expression | None = None

    properties: _GeneralPropertiesHelper = Field(default_factory=_GeneralPropertiesHelper)


class LineProperties(LayoutNode):
    class _LinePropertiesHelper(LayoutNode):
        lineColor: Expression | None = None
        roundEdge: Expression | None = None
        transparency: Expression | None = None
        weight: Expression | None = None

    properties: _LinePropertiesHelper = Field(default_factory=_LinePropertiesHelper)


class RotationProperties(LayoutNode):
    class _RotationPropertiesHelper(LayoutNode):
        angle: Expression | None = None

    properties: _RotationPropertiesHelper = Field(default_factory=_RotationPropertiesHelper)


class BasicShapeProperties(LayoutNode):
    fill: list[FillProperties] = Field(default_factory=lambda: [FillProperties()])
    general: list[GeneralProperties] = Field(default_factory=lambda: [GeneralProperties()])
    line: list[LineProperties] = Field(default_factory=lambda: [LineProperties()])
    rotation: list[RotationProperties] = Field(default_factory=lambda: [RotationProperties()])


class BasicShape(BaseVisual):
    visualType: str = "basicShape"
    model_config = ConfigDict(extra="forbid")

    drillFilterOtherVisuals: bool = True
    objects: BasicShapeProperties = Field(default_factory=BasicShapeProperties)



import shutil
from typing import TYPE_CHECKING

from pbi_core.logging import get_logger
from pbi_core.report.base import BaseReport
from pbi_core.static_files import StaticFiles

logger = get_logger()

if TYPE_CHECKING:
    from _typeshed import StrPath


class LocalStaticReport(BaseReport):
    """An instance of a PowerBI report from a local PBIX file.

    Args:
        static_files (StaticElements): An instance of all the static files (except DataModel) in the PBIX file

    Examples:
        ```python
        from pbi_core import LocalReport

        report = LocalReport.load_pbix("example.pbix")
        report.save_pbix("example_out.pbix")
        ```

    """

    _source_path: "StrPath"
    """Since this class doesn't load the full PBIX, we need to keep track of the source path for saving later"""
    static_files: StaticFiles
    """Classes representing the static design portions of the PBIX report"""

    def __init__(self, static_files: StaticFiles, source_path: "StrPath") -> None:
        self.static_files = static_files
        self._source_path = source_path

    @staticmethod
    def load_pbix(path: "StrPath") -> "LocalStaticReport":
        logger.info("Loading PBIX Static Files", path=path)
        static_files = StaticFiles.load_pbix(path)
        return LocalStaticReport(static_files=static_files, source_path=path)

    def save_pbix(self, path: "StrPath") -> None:
        """Creates a new PBIX with the information in this class to the given path.

        Examples:
            ```python

               from pbi_core import LocalReport

               report = LocalReport.load_pbix("example.pbix")
            ```
               report.save_pbix("example_out.pbix")

        Args:
            path (StrPath): the path (relative or absolute) to save the PBIX to
            sync_ssas_changes (bool, optional): whether to sync changes made in the SSAS model back to the PBIX file

        """
        shutil.copy(self._source_path, path)
        self.static_files.save_pbix(path)



from pydantic import ConfigDict, Field

from pbi_core.static_files.layout._base_node import LayoutNode

from .base import BaseVisual, ColumnProperty
from .properties.base import Expression


class GeneralProperties(LayoutNode):
    class _GeneralPropertiesHelper(LayoutNode):
        imageUrl: Expression | None = None

    properties: _GeneralPropertiesHelper = Field(default_factory=_GeneralPropertiesHelper)


class ImageScalingProperties(LayoutNode):
    class _ImageScalingPropertiesHelper(LayoutNode):
        imageScalingType: Expression | None = None

    properties: _ImageScalingPropertiesHelper = Field(default_factory=_ImageScalingPropertiesHelper)


class ImageProperties(LayoutNode):
    general: list[GeneralProperties] = Field(default_factory=list[GeneralProperties()])
    imageScaling: list[ImageScalingProperties] = Field(default_factory=list[ImageScalingProperties()])


class Image(BaseVisual):
    visualType: str = "image"
    model_config = ConfigDict(extra="forbid")

    columnProperties: dict[str, ColumnProperty] | None = None
    drillFilterOtherVisuals: bool = True
    objects: ImageProperties = Field(default_factory=ImageProperties)



from typing import TYPE_CHECKING, Literal, overload

from pbi_core.logging import get_logger
from pbi_core.ssas.model_tables.column import Column
from pbi_core.ssas.model_tables.measure import Measure
from pbi_core.ssas.server import LocalTabularModel
from pbi_core.static_files import StaticFiles

from .ssas import LocalSsasReport
from .static_files import LocalStaticReport

logger = get_logger()

if TYPE_CHECKING:
    from _typeshed import StrPath

    from pbi_core.ssas.model_tables.table import Table


class LocalReport(LocalSsasReport, LocalStaticReport):
    """An instance of a PowerBI report from a local PBIX file.

    Args:
        static_files (StaticElements): An instance of all the static files (except DataModel) in the PBIX file

    Examples:
        ```python
        from pbi_core import LocalReport

        report = LocalReport.load_pbix("example.pbix")
        report.save_pbix("example_out.pbix")
        ```

    """

    def __init__(self, ssas: LocalTabularModel, static_files: StaticFiles) -> None:
        self.ssas = ssas
        self.static_files = static_files

    @overload
    @staticmethod
    def load_pbix(
        path: "StrPath",
        *,
        kill_ssas_on_exit: bool = True,
        load_ssas: Literal[True] = True,
        load_static_files: Literal[True] = True,
    ) -> "LocalReport": ...

    @overload
    @staticmethod
    def load_pbix(  # pyright: ignore[reportOverlappingOverload]
        path: "StrPath",
        *,
        kill_ssas_on_exit: bool = True,
        load_ssas: Literal[False] = False,
        load_static_files: Literal[True] = True,
    ) -> "LocalStaticReport": ...

    @overload
    @staticmethod
    def load_pbix(
        path: "StrPath",
        *,
        kill_ssas_on_exit: bool = True,
        load_ssas: Literal[True] = True,
        load_static_files: Literal[False] = False,
    ) -> "LocalSsasReport": ...

    @overload
    @staticmethod
    def load_pbix(
        path: "StrPath",
        *,
        kill_ssas_on_exit: bool = True,
        load_ssas: bool = True,
        load_static_files: bool = False,
    ) -> "LocalReport": ...

    # If the above overloads are not used, this will be the default implementation, assuming both sides are loaded

    @staticmethod
    def load_pbix(  # pyright: ignore[reportIncompatibleMethodOverride]
        path: "StrPath",
        *,
        kill_ssas_on_exit: bool = True,
        load_ssas: bool = True,
        load_static_files: bool = True,
    ) -> "LocalReport | LocalSsasReport | LocalStaticReport":
        """Creates a ``LocalReport`` instance from a PBIX file.

        Args:
            path (StrPath): The absolute or local path to the PBIX report
            kill_ssas_on_exit (bool, optional): The LocalReport object depends on a ``msmdsrv.exe`` process that is
                independent of the Python session process. If this function creates a new ``msmdsrv.exe`` instance
                and kill_ssas_on_exit is true, the process will be killed on exit.
            load_ssas (bool, optional): Indicates if the report should load any objects within the report.ssas
                attribute
            load_static_files (bool, optional): Indicates if the report should load any objects within the
                report.static_files attribute

        Raises:
            ValueError: Occurs when neither the load_ssas nor load_static_files parameters are true,
                so there's nothing to load.

        Examples:
            ```python

               from pbi_core import LocalReport

               report = LocalReport.load_pbix("example.pbix")
            ```

        Returns:
                LocalReport: the local PBIX class

        """
        logger.info(
            "Loading PBIX report",
            path=path,
            load_ssas=load_ssas,
            load_static_files=load_static_files,
        )
        if load_ssas & load_static_files:
            ssas_content = LocalSsasReport.load_pbix(path, kill_ssas_on_exit=kill_ssas_on_exit)
            static_content = LocalStaticReport.load_pbix(path)
            logger.info("Loaded PBIX report", components="ssas+static")
            return LocalReport(ssas=ssas_content.ssas, static_files=static_content.static_files)
        if load_ssas:
            logger.info("Loaded PBIX report", components="ssas")
            return LocalSsasReport.load_pbix(path, kill_ssas_on_exit=kill_ssas_on_exit)
        if load_static_files:
            logger.info("Loaded PBIX report", components="static")
            return LocalStaticReport.load_pbix(path)
        msg = "At least one of load_ssas or load_static_files must be True"
        raise ValueError(msg)

    def save_pbix(self, path: "StrPath", *, sync_ssas_changes: bool = True) -> None:
        """Creates a new PBIX with the information in this class to the given path.

        Examples:
            ```python

               from pbi_core import LocalReport

               report = LocalReport.load_pbix("example.pbix")
            ```
               report.save_pbix("example_out.pbix")

        Args:
            path (StrPath): the path (relative or absolute) to save the PBIX to
            sync_ssas_changes (bool, optional): whether to sync changes made in the SSAS model back to the PBIX file

        """
        logger.info("Saving PBIX report", path=path, sync_ssas_changes=sync_ssas_changes)
        # Dev note: DO NOT call save_pbix of LocalSsasReport or LocalStaticReport, as they are implemented assuming the
        # other doesn't exist, causing changes to be overwritten by them copying the source file
        if sync_ssas_changes:
            self.ssas.sync_to()
        self.ssas.save_pbix(path)
        self.static_files.save_pbix(path)
        logger.info("Saved PBIX report", path=path)

    def cleanse_ssas_model(self) -> None:
        """Removes all unused tables, columns, and measures in an SSAS model.

        1. Uses the layout to identify all Measures and Columns being used by the report visuals and filters.
        2. Uses SSAS relationships to identify additional columns and tables used for cross-table filtering.
        3. Traces calculation dependencies (on measures and calculated columns) to identify measures and columns used
             to create report fields
        4. Removes any measure/column that is:
            1. Not in results of 1-3
            2. Not part of a system table
        5. Removes any table that has no column/measure used in the report and no active relationship with a reporting
            table
        """
        report_references = self.static_files.layout.get_ssas_elements()
        model_values = (
            [x.to_model(self.ssas) for x in report_references]
            + [relationship.to_column() for relationship in self.ssas.relationships]
            + [relationship.from_column() for relationship in self.ssas.relationships]
        )
        ret: list[Measure | Column] = []
        for val in model_values:
            ret.append(val)
            ret.extend(val.parents(recursive=True))

        used_tables = {x.table() for x in ret}
        used_measures = {
            x
            for x in ret
            if isinstance(
                x,
                Measure,
            )
        }
        used_columns = {x for x in ret if isinstance(x, Column)}

        # In the examples I've seen, there's a table named "DateTableTemplate_<UUID>" that doesn't seem used,
        # but breaks the system when removed
        tables_to_drop = {t for t in self.ssas.tables if t not in used_tables and not t.is_private}
        columns_to_drop = {
            c
            for c in self.ssas.columns
            if c not in used_columns
            and not c.table().name.startswith("DateTableTemplate")
            and c.table() not in tables_to_drop
        }
        logger.info(
            "Dropping unused SSAS elements",
            tables_to_drop=len(tables_to_drop),
            columns_to_drop=len(columns_to_drop),
        )
        affected_tables: dict[Table, list[Column]] = {}
        for c in columns_to_drop:
            t = c.table()
            if t not in tables_to_drop:
                affected_tables.setdefault(t, []).append(c)

        measures_to_drop = {m for m in self.ssas.measures if m not in used_measures}
        # TODO: convert to batch deletion
        for t in tables_to_drop:
            t.delete()
        for c in columns_to_drop:
            c.delete()
        for m in measures_to_drop:
            m.delete()
        for affected_table, table_columns_to_drop in affected_tables.items():
            for partition in affected_table.partitions():
                partition.remove_columns(table_columns_to_drop)
        logger.info(
            "SSAS model cleansed",
            tables_dropped=len(tables_to_drop),
            columns_dropped=len(columns_to_drop),
            measures_dropped=len(measures_to_drop),
        )



import datetime
import json
import math
import textwrap
import threading
import time
from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any

import jinja2
from pbi_pyadomd import Connection, Reader

from pbi_core.logging import get_logger
from pbi_core.ssas.server.tabular_model.tabular_model import BaseTabularModel
from pbi_core.ssas.server.trace.trace_enums import TraceEvents

logger = get_logger()
TRACE_DIR = Path(__file__).parent / "templates"
TRACE_TEMPLATES: dict[str, jinja2.Template] = {
    f.name: jinja2.Template(f.read_text()) for f in TRACE_DIR.iterdir() if f.is_file()
}

TRACE_DEFAULT_EVENTS = (
    TraceEvents.DISCOVER_BEGIN,
    TraceEvents.COMMAND_BEGIN,
    TraceEvents.QUERY_END,
    TraceEvents.QUERY_SUBCUBE,
    TraceEvents.VERTIPAQ_SE_QUERY_BEGIN,
    TraceEvents.VERTIPAQ_SE_QUERY_END,
    TraceEvents.VERTIPAQ_SE_QUERY_CACHE_MATCH,
    TraceEvents.VERTIPAQ_SE_QUERY_CACHE_MISS,
    TraceEvents.AGGREGATE_TABLE_REWRITE_QUERY,
    TraceEvents.DIRECT_QUERY_END,
    TraceEvents.QUERY_BEGIN,
    TraceEvents.EXECUTION_METRICS,
)


def pretty_size(n: int) -> str:
    """Convert a number of bytes into a human-readable format."""
    units = ["B"] + [f"{p}iB" for p in "KMGTPEZY"]

    prefix = int(math.log(max(n, 1), 1024))
    prefix = min(prefix, len(units) - 1)
    pretty_n = round(n / 1024**prefix, 1)
    return f"{pretty_n} {units[prefix]}"


@dataclass
class ThreadResult:
    command: str
    rows_returned: int
    error: Exception | None = None

    def get_performance(self, trace_records: list[dict[str, Any]]) -> "Performance":
        event_info = {}
        for record in trace_records:
            event_info[record["EventClass"]] = record

        execution_metrics = json.loads(event_info["EXECUTION_METRICS"]["TextData"])
        command_text = event_info["QUERY_END"]["TextData"]
        start_datetime = datetime.datetime.strptime(execution_metrics["timeStart"], "%Y-%m-%dT%H:%M:%S.%f%z")
        end_datetime = datetime.datetime.strptime(execution_metrics["timeEnd"], "%Y-%m-%dT%H:%M:%S.%f%z")
        total_duration = execution_metrics["durationMs"]
        total_cpu_time = execution_metrics["totalCpuTimeMs"]
        query_cpu_time = execution_metrics["queryProcessingCpuTimeMs"]
        vertipaq_cpu_time = execution_metrics["vertipaqJobCpuTimeMs"]
        execution_delay = execution_metrics["executionDelayMs"]
        approximate_peak_consumption_kb = execution_metrics["approximatePeakMemConsumptionKB"]

        for record in trace_records:
            if record.get("EventClass") == "QUERY_END":
                return Performance(
                    command_text=command_text,
                    start_datetime=start_datetime,
                    end_datetime=end_datetime,
                    total_duration=total_duration,
                    total_cpu_time=total_cpu_time,
                    query_cpu_time=query_cpu_time,
                    vertipaq_cpu_time=vertipaq_cpu_time,
                    execution_delay=execution_delay,
                    approximate_peak_consumption_kb=approximate_peak_consumption_kb,
                    rows_returned=self.rows_returned,
                    full_trace_log=trace_records,
                )
        msg = f"Command '{self.command}' not found in trace records."
        raise ValueError(msg)


@dataclass
class Performance:
    command_text: str
    """The text of the command being executed"""
    start_datetime: datetime.datetime
    """When the query started"""
    end_datetime: datetime.datetime
    """When the query ended"""
    query_cpu_time: int
    """Total Query CPU Time (i.e. the sum of the time each core involved was used)"""
    total_duration: int
    """Total duration is the total time in milliseconds"""
    total_cpu_time: int
    """Total CPU Time (i.e. the sum of the time each core involved was used)"""
    vertipaq_cpu_time: int
    """CPU Time spent by the vertipaq engine"""
    execution_delay: int
    """Delay between the query start and processing due to other competing queries"""
    approximate_peak_consumption_kb: int
    """The max memory used by the query throughout its execution"""
    rows_returned: int
    """The number of rows returned by the query"""
    full_trace_log: list[dict[str, Any]]

    def __repr__(self) -> str:
        total_duration = round(self.total_duration / 1000, 2)
        total_cpu_time = round(self.total_cpu_time / 1000, 2)
        approximate_peak_consumption = pretty_size(self.approximate_peak_consumption_kb * 1024)
        return f"Performance(rows={self.rows_returned}, total_duration={total_duration}, total_cpu_time={total_cpu_time}, peak_consumption={approximate_peak_consumption}"  # noqa: E501

    def pprint(self) -> str:
        """Pretty print the performance metrics."""
        command_text = textwrap.indent(self.command_text, " " * 8)
        return f"""Performance(
    Command:

{command_text}

    Start Time: {self.start_datetime.isoformat()}
    End Time: {self.end_datetime.isoformat()}
    Total Duration: {self.total_duration} ms
    Total CPU Time: {self.total_cpu_time} ms
    Query CPU Time: {self.query_cpu_time} ms
    Vertipaq CPU Time: {self.vertipaq_cpu_time} ms
    Execution Delay: {self.execution_delay} ms
    Approximate Peak Consumption: {pretty_size(self.approximate_peak_consumption_kb * 1024)}
    Rows Returned: {self.rows_returned}
)"""


class SubscriberState(Enum):
    INITIALIZED = "initialized"
    SUBSCRIBED = "subscribed"


class Subscriber:
    _state: SubscriberState = SubscriberState.INITIALIZED
    """The state is only used to manage the _pinger method"""

    trace_records: dict[str, list[dict[str, Any]]]
    reader: Reader

    def __init__(self, subscription_create_command: str, conn: Connection, events: Iterable[TraceEvents]) -> None:
        """Initializes the Subscriber in Python and SSAS.

        Note: Without the self.pinger on a separate doing a trivial ping, the subscription occasionally
        (10% in tests) takes forever (1-5 minutes vs 1-2 seconds) to be created. Although hopefully
        we solve the root cause, this workaround appears to solve the issue for now.
        """
        ping_thread = threading.Thread(target=self.pinger, args=(conn.clone(),), daemon=True)
        ping_thread.start()
        self.reader = conn.execute_dax(
            subscription_create_command,
        )  # occasionally seems to take forever (1-5 minutes vs 1-2 seconds)
        self._state = SubscriberState.SUBSCRIBED

        self.events = events
        self.trace_records = {}
        self.command_request_ids: dict[str, str] = {}
        self.thread = threading.Thread(target=self.poll_cursor, daemon=True)
        self.thread.start()

    def pinger(self, conn: Connection) -> None:
        """Only used to remind the SSAS that there's a Subscription waiting to be created."""
        conn.open()
        while self._state != SubscriberState.SUBSCRIBED:
            time.sleep(1)
            conn.execute_dax(
                "EVALUATE {1}",
            ).close()  # simple ping to keep the connection alive. We don't actually care about the output
        conn.close()

    def poll_cursor(self) -> None:
        event_mapping = {e.value: e.name for e in self.events}
        for record in self.reader.fetch_stream():
            if "EventClass" in record:
                record["EventClass"] = event_mapping[record["EventClass"]]
            self.trace_records.setdefault(record["RequestID"], []).append(record)
            if "TextData" in record:
                self.command_request_ids[record["TextData"]] = record["RequestID"]

    def kill_polling(self) -> None:
        if self.thread.is_alive():
            self.thread.join(timeout=1)

    def __repr__(self) -> str:
        return f"Subscriber(records={len(self.trace_records)})"


class PerformanceTrace:
    events: Iterable[TraceEvents]
    db: BaseTabularModel
    commands: list[str]
    trace_create_command: str
    subscription_create_command: str
    trace_delete_command: str
    clear_cache_command: str

    def __init__(
        self,
        db: BaseTabularModel,
        commands: list[str],
        events: Iterable[TraceEvents] = TRACE_DEFAULT_EVENTS,
    ) -> None:
        self.events = events
        self.db: BaseTabularModel = db
        self.commands = [x.replace("\r\n", "\n") for x in commands]

        next_day = (datetime.datetime.now(datetime.UTC) + datetime.timedelta(days=1)).strftime("%Y-%m-%dT%H:%M:%SZ")
        trace_name = f"pbi_core_{next_day.replace(':', '_')}"
        subscription_name = f"pbi_core_subscription_{next_day.replace(':', '_')}"
        self.trace_create_command = TRACE_TEMPLATES["trace_create.xml"].render(
            trace_name=trace_name,
            stop_time=next_day,
            events=self.events,
        )
        self.subscription_create_command = TRACE_TEMPLATES["subscription_create.xml"].render(
            trace_name=trace_name,
            subscription_name=subscription_name,
        )
        self.trace_delete_command = TRACE_TEMPLATES["trace_delete.xml"].render(
            trace_name=trace_name,
        )
        self.clear_cache_command = TRACE_TEMPLATES["clear_cache.xml"].render(database_name=self.db.db_name)

    def initialize_tracing(self) -> "PerformanceTrace":
        logger.info("Beginning trace")
        self.db.server.query_xml(self.trace_create_command)
        self.subscriber = Subscriber(
            self.subscription_create_command,
            self.get_conn(),
            self.events,
        )
        return self

    def terminate_tracing(self) -> None:
        logger.info("Terminating trace")
        with self.db.server.conn(db_name=self.db.db_name) as conn:
            conn.execute_xml(self.trace_delete_command)

    def get_performance(self, *, clear_cache: bool = False) -> list[Performance]:
        def thread_func(command: str) -> ThreadResult:
            try:
                with self.get_conn() as conn:
                    reader = conn.execute_dax(command)
                    rows_returned = len(reader.fetch_many(limit=500))

                    # See note in Reader.fetch_stream() documentation for why we do this
                    reader.close()

            except ValueError as e:
                return ThreadResult(
                    command=command,
                    rows_returned=-1,
                    error=e,
                )
            # The limit of 500 is to mimic the behavior of PowerBI, which returns by defult 500 rows
            return ThreadResult(
                command=command,
                rows_returned=rows_returned,
            )

        self.initialize_tracing()

        if clear_cache:
            logger.info("Clearing cache before performance testing")
            with self.db.server.conn(db_name=self.db.db_name) as conn:
                conn.execute_xml(self.clear_cache_command)

        with ThreadPoolExecutor() as dax_executor:
            logger.info("Running DAX commands")
            command_results = list(dax_executor.map(thread_func, self.commands))

        missing_subscription_records = command_results
        for _ in range(5):
            # Testing for 15 seconds to check that all commands have an entry in the trace
            new_missing_subscription_records = []
            for command in missing_subscription_records:
                request_id = self.subscriber.command_request_ids.get(command.command)
                if request_id is None:
                    new_missing_subscription_records.append(command)
                    continue
                command_events = self.subscriber.trace_records.get(request_id, [])
                if not any(x["EventClass"] == "QUERY_END" for x in command_events):
                    new_missing_subscription_records.append(command)

            if not new_missing_subscription_records:
                break
            time.sleep(1)
            missing_subscription_records = new_missing_subscription_records
        else:
            missing_commands = ", ".join(cmd.command for cmd in missing_subscription_records)
            msg = f"Some commands did not have trace records: {missing_commands}"
            raise ValueError(msg)

        self.subscriber.kill_polling()
        self.terminate_tracing()

        return [
            command_result.get_performance(
                self.subscriber.trace_records[self.subscriber.command_request_ids[command_result.command]],
            )
            for command_result in command_results
        ]

    def get_conn(self) -> Connection:
        return self.db.server.conn(db_name=self.db.db_name).open()



from pbi_core.logging import get_logger
from pbi_core.report.local.main import LocalReport
from pbi_core.ssas.model_tables.enums.enums import DataType
from pbi_core.ssas.server.tabular_model.tabular_model import BaseTabularModel

from .get_static_elements import get_static_elements
from .text_elements import TextElement, TextElements

logger = get_logger()


def get_ssas_elements(server: BaseTabularModel) -> TextElements:
    text_elements = []
    for m in server.measures:
        text_elements.append(
            TextElement(
                category="Measure",
                source="ssas",
                xpath=["measures", m.id],
                field="name",
                text=m.name,
            ),
        )
        if m.data_type == DataType.STRING and isinstance(m.expression, str):
            text_elements.append(
                TextElement(
                    category="Measure",
                    source="ssas",
                    xpath=["measures", m.id],
                    field="expression",
                    text=m.expression,
                ),
            )
        if isinstance(m.format_string, str):
            text_elements.append(
                TextElement(
                    category="Measure",
                    source="ssas",
                    xpath=["measures", m.id],
                    field="format_string",
                    text=m.format_string,
                ),
            )
        if isinstance(m.description, str):
            text_elements.append(
                TextElement(
                    category="Measure",
                    source="ssas",
                    xpath=["measures", m.id],
                    field="description",
                    text=m.description,
                ),
            )
    for c in server.columns:
        if isinstance(c.explicit_name, str):
            text_elements.append(
                TextElement(
                    category="Column",
                    source="ssas",
                    xpath=["columns", c.id],
                    field="explicit_name",
                    text=c.explicit_name,
                ),
            )
        if isinstance(c.description, str):
            text_elements.append(
                TextElement(
                    category="Column",
                    source="ssas",
                    xpath=["columns", c.id],
                    field="description",
                    text=c.description,
                ),
            )
    return TextElements(text_elements=text_elements)


def get_source_text_elements(report: LocalReport) -> TextElements:
    static_elements = get_static_elements(report.static_files.layout)
    ssas_elements = get_ssas_elements(report.ssas)
    return TextElements(text_elements=static_elements.text_elements + ssas_elements.text_elements)



from pbi_core.logging import get_logger
from pbi_core.pydantic.main import BaseValidation
from pbi_core.static_files import Layout
from pbi_core.static_files.layout.filters import Filter
from pbi_core.static_files.layout.sources.literal import LiteralSource
from pbi_core.static_files.layout.sources.paragraphs import TextRun
from pbi_core.static_files.layout.visuals.base import PropertyDef
from pbi_core.static_files.layout.visuals.properties.base import LiteralExpression
from pbi_core.static_files.layout.visuals.text_box import TextBox

from .text_elements import TextElement, TextElements

logger = get_logger()


def _parse_viz_config(config: BaseValidation | dict | None) -> list[LiteralSource]:
    if config is None:
        return []
    ret: list[LiteralSource] = []
    if isinstance(config, dict):
        for value in config.values():
            for element in value:
                assert isinstance(element, PropertyDef)
                assert isinstance(element.properties, dict)
                ret.extend(
                    field_value.expr
                    for field_value in element.properties.values()
                    if isinstance(field_value, LiteralExpression) and isinstance(field_value.expr.value(), str)
                )
    else:
        for field_name in config.__pydantic_fields__:
            value: list[BaseValidation] = getattr(config, field_name)
            for element in value:
                properties: BaseValidation = element.properties  # type: ignore reportAttributeAccessIssue
                for prop_name in properties.__pydantic_fields__:
                    prop_value = getattr(properties, prop_name)
                    if isinstance(prop_value, LiteralExpression) and isinstance(prop_value.expr.value(), str):
                        ret.append(prop_value.expr)
    return ret


def _get_text_runs(viz: TextBox) -> list[TextRun]:
    ret = []
    for property_list in viz.objects.general:
        for paragraph in property_list.properties.paragraphs or []:
            ret.extend(run for run in paragraph.textRuns)
    return ret


def get_static_elements(layout: Layout) -> TextElements:
    """Retrieves all static elements of a report Layout.

    The static elements in the report are:
    1. Section names
    2. Filter names
    3. Visual config names such as title, header, etc.
    """
    elements = []

    elements.extend(
        TextElement(
            category="Filter",
            source="layout",
            xpath=f.get_xpath(layout),
            field="displayName",
            text=f.displayName,
        )
        for f in layout.find_all(Filter)
        if f.displayName is not None
    )

    for section in layout.sections:
        elements.append(
            TextElement(
                category="Section",
                source="layout",
                xpath=section.get_xpath(layout),
                field="displayName",
                text=section.displayName,
            ),
        )

        for visual_container in section.visualContainers:
            for visual in visual_container.get_visuals():
                if isinstance(visual, TextBox):
                    text_runs = _get_text_runs(visual)
                    elements.extend(
                        TextElement(
                            category="Visual",
                            source="layout",
                            xpath=run.get_xpath(layout),
                            field="value",
                            text=run.value,
                        )
                        for run in text_runs
                        if isinstance(run.value, str)
                    )
                for config_obj in [visual.vcObjects, visual.objects]:
                    text_config = _parse_viz_config(config_obj)
                    for prop in text_config:
                        val = prop.value()
                        assert isinstance(
                            val,
                            str,
                        )  # technically done before, but the type checker doesn't remember that
                        elements.append(
                            TextElement(
                                category="Visual",
                                source="layout",
                                xpath=prop.get_xpath(layout),
                                field="expr",
                                text=val,
                            ),
                        )

    return TextElements(elements)



import datetime
from typing import TYPE_CHECKING

from pydantic import PrivateAttr

from pbi_core.ssas.model_tables.base import SsasEditableRecord
from pbi_core.ssas.server._commands import BaseCommands
from pbi_core.ssas.server.utils import SsasCommands

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.table import Table


class CalculationGroup(SsasEditableRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/ed9dcbcf-9910-455f-abc4-13c575157cfb)
    """

    description: str
    precedence: int
    table_id: int

    modified_time: datetime.datetime

    _commands: BaseCommands = PrivateAttr(default_factory=lambda: SsasCommands.calculation_group)

    def table(self) -> "Table":
        return self.tabular_model.tables.find(self.table_id)



from pydantic import ConfigDict, Field

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector

from .base import BaseVisual, ColumnProperty
from .properties.base import Expression


class AnomalyDetectionProperties(LayoutNode):
    class _AnomalyDetectionPropertiesHelper(LayoutNode):
        confidenceBandColor: Expression | None = None
        displayName: Expression | None = None
        explainBy: Expression | None = None
        markerColor: Expression | None = None
        markerShape: Expression | None = None
        show: Expression | None = None
        transform: Expression | None = None
        transparency: Expression | None = None

    properties: _AnomalyDetectionPropertiesHelper = Field(default_factory=_AnomalyDetectionPropertiesHelper)
    selector: Selector | None = None


class CategoryAxisProperties(LayoutNode):
    class _CategoryAxisPropertiesHelper(LayoutNode):
        axisType: Expression | None = None
        concatenateLabels: Expression | None = None
        end: Expression | None = None
        fontFamily: Expression | None = None
        fontSize: Expression | None = None
        gridlineShow: Expression | None = None
        labelColor: Expression | None = None
        maxMarginFactor: Expression | None = None
        show: Expression | None = None
        showAxisTitle: Expression | None = None
        start: Expression | None = None
        titleColor: Expression | None = None
        titleFontFamily: Expression | None = None
        titleFontSize: Expression | None = None

    properties: _CategoryAxisPropertiesHelper = Field(default_factory=_CategoryAxisPropertiesHelper)
    selector: Selector | None = None


class DataPointProperties(LayoutNode):
    class _DataPointPropertiesHelper(LayoutNode):
        fill: Expression | None = None
        showAllDataPoints: Expression | None = None

    properties: _DataPointPropertiesHelper = Field(default_factory=_DataPointPropertiesHelper)
    selector: Selector | None = None


class ForecastProperties(LayoutNode):
    class _ForecastPropertiesHelper(LayoutNode):
        show: Expression | None = None
        displayName: Expression | None = None
        lineColor: Expression | None = None
        transform: Expression | None = None

    properties: _ForecastPropertiesHelper = Field(default_factory=_ForecastPropertiesHelper)
    selector: Selector | None = None


class GeneralProperties(LayoutNode):
    class _GeneralPropertiesHelper(LayoutNode):
        responsive: Expression | None = None

    properties: _GeneralPropertiesHelper = Field(default_factory=_GeneralPropertiesHelper)


class LabelsProperties(LayoutNode):
    class _LabelsPropertiesHelper(LayoutNode):
        color: Expression | None = None
        labelPosition: Expression | None = None
        fontFamily: Expression | None = None
        fontSize: Expression | None = None
        labelDensity: Expression | None = None
        show: Expression | None = None
        showAll: Expression | None = None
        showSeries: Expression | None = None

    properties: _LabelsPropertiesHelper = Field(default_factory=_LabelsPropertiesHelper)
    selector: Selector | None = None


class LegendProperties(LayoutNode):
    class _LegendPropertiesHelper(LayoutNode):
        defaultToCircle: Expression | None = None
        fontSize: Expression | None = None
        labelColor: Expression | None = None
        legendMarkerRendering: Expression | None = None
        position: Expression | None = None
        show: Expression | None = None
        showTitle: Expression | None = None
        titleText: Expression | None = None

    properties: _LegendPropertiesHelper = Field(default_factory=_LegendPropertiesHelper)


class LineStylesProperties(LayoutNode):
    class _LineStylesPropertiesHelper(LayoutNode):
        lineStyle: Expression | None = None
        markerColor: Expression | None = None
        markerShape: Expression | None = None
        markerSize: Expression | None = None
        showMarker: Expression | None = None
        showSeries: Expression | None = None
        stepped: Expression | None = None
        strokeLineJoin: Expression | None = None
        strokeWidth: Expression | None = None

    properties: _LineStylesPropertiesHelper = Field(default_factory=_LineStylesPropertiesHelper)
    selector: Selector | None = None


class PlotAreaProperties(LayoutNode):
    class _PlotAreaPropertiesHelper(LayoutNode):
        transparency: Expression | None = None

    properties: _PlotAreaPropertiesHelper = Field(default_factory=_PlotAreaPropertiesHelper)


class TrendProperties(LayoutNode):
    class _TrendPropertiesHelper(LayoutNode):
        displayName: Expression | None = None
        lineColor: Expression | None = None
        show: Expression | None = None

    properties: _TrendPropertiesHelper = Field(default_factory=_TrendPropertiesHelper)


class ValueAxisProperties(LayoutNode):
    class _ValueAxisPropertiesHelper(LayoutNode):
        axisScale: Expression | None = None
        end: Expression | None = None
        fontFamily: Expression | None = None
        fontSize: Expression | None = None
        gridlineColor: Expression | None = None
        gridlineShow: Expression | None = None
        gridlineStyle: Expression | None = None
        gridlineThickness: Expression | None = None
        labelColor: Expression | None = None
        labelDensity: Expression | None = None
        labelDisplayUnits: Expression | None = None
        position: Expression | None = None
        show: Expression | None = None
        showAxisTitle: Expression | None = None
        start: Expression | None = None
        titleFontFamily: Expression | None = None
        titleText: Expression | None = None

    properties: _ValueAxisPropertiesHelper = Field(default_factory=_ValueAxisPropertiesHelper)


class Y1AxisReferenceLineProperties(LayoutNode):
    class _Y1AxisReferenceLinePropertiesHelper(LayoutNode):
        displayName: Expression | None = None
        position: Expression | None = None
        show: Expression | None = None
        style: Expression | None = None

    properties: _Y1AxisReferenceLinePropertiesHelper = Field(default_factory=_Y1AxisReferenceLinePropertiesHelper)
    selector: Selector | None = None


class Y2AxisProperties(LayoutNode):
    class _Y2AxisPropertiesHelper(LayoutNode):
        show: Expression | None = None

    properties: _Y2AxisPropertiesHelper = Field(default_factory=_Y2AxisPropertiesHelper)


class ZoomProperties(LayoutNode):
    class _ZoomPropertiesHelper(LayoutNode):
        show: Expression | None = None
        categoryMax: Expression | None = None
        categoryMin: Expression | None = None
        valueMax: Expression | None = None
        valueMin: Expression | None = None

    properties: _ZoomPropertiesHelper = Field(default_factory=_ZoomPropertiesHelper)


class LineChartProperties(LayoutNode):
    anomalyDetection: list[AnomalyDetectionProperties] = Field(
        default_factory=lambda: [AnomalyDetectionProperties()],
    )
    categoryAxis: list[CategoryAxisProperties] = Field(default_factory=lambda: [CategoryAxisProperties()])
    dataPoint: list[DataPointProperties] = Field(default_factory=lambda: [DataPointProperties()])
    forecast: list[ForecastProperties] = Field(default_factory=lambda: [ForecastProperties()])
    general: list[GeneralProperties] = Field(default_factory=lambda: [GeneralProperties()])
    labels: list[LabelsProperties] = Field(default_factory=lambda: [LabelsProperties()])
    legend: list[LegendProperties] = Field(default_factory=lambda: [LegendProperties()])
    lineStyles: list[LineStylesProperties] = Field(default_factory=lambda: [LineStylesProperties()])
    plotArea: list[PlotAreaProperties] = Field(default_factory=lambda: [PlotAreaProperties()])
    trend: list[TrendProperties] = Field(default_factory=lambda: [TrendProperties()])
    valueAxis: list[ValueAxisProperties] = Field(default_factory=lambda: [ValueAxisProperties()])
    zoom: list[ZoomProperties] = Field(default_factory=lambda: [ZoomProperties()])
    y1AxisReferenceLine: list[Y1AxisReferenceLineProperties] = Field(
        default_factory=lambda: [Y1AxisReferenceLineProperties()],
    )
    y2Axis: list[Y2AxisProperties] = Field(default_factory=lambda: [Y2AxisProperties()])


class LineChart(BaseVisual):
    visualType: str = "lineChart"
    model_config = ConfigDict(extra="forbid")

    columnProperties: dict[str, ColumnProperty] | None = None
    drillFilterOtherVisuals: bool = True
    objects: LineChartProperties = Field(default_factory=LineChartProperties)



from typing import Any

from pydantic import ConfigDict, Field

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector
from pbi_core.static_files.layout.sources.paragraphs import Paragraph

from .base import BaseVisual
from .properties.base import Expression


class GeneralProperties(LayoutNode):
    class _GeneralPropertiesHelper(LayoutNode):
        paragraphs: list[Paragraph] | None = None
        responsive: Expression | None = None

    properties: _GeneralPropertiesHelper = Field(default_factory=_GeneralPropertiesHelper)


class ValueProperties(LayoutNode):
    class _ValuePropertiesHelper(LayoutNode):
        class _ValuePropertiesExpr(LayoutNode):
            context: Any | None = None  # TODO: should be Source, but causes circular import issues with Subquery
            expr: Any | None = None  # TODO: should be Source, but causes circular import issues with Subquery
            value: Any | None = None  # TODO: should be Source, but causes circular import issues with Subquery
            propertyDefinitionKind: str | None = None

        expr: _ValuePropertiesExpr = Field(default_factory=_ValuePropertiesExpr)
        formatString: Expression | None = None

    properties: _ValuePropertiesHelper = Field(default_factory=_ValuePropertiesHelper)
    selector: Selector | None = None


class TextBoxProperties(LayoutNode):
    general: list[GeneralProperties] = Field(default_factory=lambda: [GeneralProperties()])
    values: list[ValueProperties] = Field(default_factory=lambda: [ValueProperties()])


class TextBox(BaseVisual):
    visualType: str = "textbox"
    model_config = ConfigDict(extra="forbid")

    drillFilterOtherVisuals: bool = True
    objects: TextBoxProperties = Field(default_factory=TextBoxProperties)



import sys

from pydantic import BaseModel, ConfigDict

ALLOW_EXTRA = "allow" if "--allow-extra" in sys.argv else "forbid"


class BaseValidation(BaseModel):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra=ALLOW_EXTRA,
        use_enum_values=False,
        json_schema_mode_override="serialization",
        validate_assignment=True,
        protected_namespaces=(),
    )

    def __format__(self, format_spec: str) -> str:
        return f"{self!s:{format_spec}}"

    def model_dump_json(self, *, indent: int = 4) -> str:  # type: ignore[override]
        # alias is required for models since the PBIX format uses the alias version of attribute names
        # exclude_unset is required to avoid writing default values that are not present in the original PBIX.
        #  This can cause issues otherwise with visual configs
        # round_trip is required to return JSON types back to strings when 
        return super().model_dump_json(by_alias=True, exclude_unset=True, indent=indent, round_trip=True)



from pydantic import Field

from pbi_core.static_files.layout._base_node import LayoutNode

from .base import Expression


class BackgroundProperties(LayoutNode):
    class _BackgroundPropertiesHelper(LayoutNode):
        color: Expression | None = None
        show: Expression | None = None
        transparency: Expression | None = None

    properties: _BackgroundPropertiesHelper = Field(default_factory=_BackgroundPropertiesHelper)


class BorderProperties(LayoutNode):
    class _BorderPropertiesHelper(LayoutNode):
        background: Expression | None = None
        color: Expression | None = None
        radius: Expression | None = None
        show: Expression | None = None
        width: Expression | None = None

    properties: _BorderPropertiesHelper = Field(default_factory=_BorderPropertiesHelper)


class DividerProperties(LayoutNode):
    class _DividerPropertiesHelper(LayoutNode):
        color: Expression | None = None
        show: Expression | None = None
        style: Expression | None = None
        width: Expression | None = None

    properties: _DividerPropertiesHelper = Field(default_factory=_DividerPropertiesHelper)


class DropShadowProperties(LayoutNode):
    class _DropShadowPropertiesHelper(LayoutNode):
        angle: Expression | None = None
        color: Expression | None = None
        position: Expression | None = None
        preset: Expression | None = None
        shadowBlur: Expression | None = None
        shadowDistance: Expression | None = None
        shadowSpread: Expression | None = None
        show: Expression | None = None
        transparency: Expression | None = None

    properties: _DropShadowPropertiesHelper = Field(default_factory=_DropShadowPropertiesHelper)


class GeneralProperties(LayoutNode):
    class _GeneralPropertiesHelper(LayoutNode):
        altText: Expression | None = None
        keepLayerOrder: Expression | None = None

    properties: _GeneralPropertiesHelper = Field(default_factory=_GeneralPropertiesHelper)


class LockAspectProperties(LayoutNode):
    class _LockAspectPropertiesHelper(LayoutNode):
        show: Expression | None = None

    properties: _LockAspectPropertiesHelper = Field(default_factory=_LockAspectPropertiesHelper)


class SpacingProperties(LayoutNode):
    class _SpacingPropertiesHelper(LayoutNode):
        customizeSpacing: Expression | None = None
        spaceBelowSubTitle: Expression | None = None
        spaceBelowTitle: Expression | None = None
        spaceBelowTitleArea: Expression | None = None

    properties: _SpacingPropertiesHelper = Field(default_factory=_SpacingPropertiesHelper)


class StylePresetProperties(LayoutNode):
    class _StylePresetPropertiesHelper(LayoutNode):
        name: Expression | None = None

    properties: _StylePresetPropertiesHelper = Field(default_factory=_StylePresetPropertiesHelper)


class SubTitleProperties(LayoutNode):
    class _SubTitlePropertiesHelper(LayoutNode):
        alignment: Expression | None = None
        bold: Expression | None = None
        fontColor: Expression | None = None
        fontFamily: Expression | None = None
        heading: Expression | None = None
        show: Expression | None = None
        text: Expression | None = None
        titleWrap: Expression | None = None

    properties: _SubTitlePropertiesHelper = Field(default_factory=_SubTitlePropertiesHelper)


class TitleProperties(LayoutNode):
    class _TitlePropertiesHelper(LayoutNode):
        alignment: Expression | None = None
        background: Expression | None = None
        fontColor: Expression | None = None
        fontFamily: Expression | None = None
        fontSize: Expression | None = None
        heading: Expression | None = None
        show: Expression | None = None
        text: Expression | None = None
        titleWrap: Expression | None = None
        underline: Expression | None = None

    properties: _TitlePropertiesHelper = Field(default_factory=_TitlePropertiesHelper)


class VisualHeaderProperties(LayoutNode):
    class _VisualHeaderPropertiesHelper(LayoutNode):
        background: Expression | None = None
        border: Expression | None = None
        foreground: Expression | None = None
        show: Expression | None = None
        showDrillDownExpandButton: Expression | None = None
        showDrillDownLevelButton: Expression | None = None
        showDrillRoleSelector: Expression | None = None
        showDrillToggleButton: Expression | None = None
        showDrillUpButton: Expression | None = None
        showFilterRestatementButton: Expression | None = None
        showFocusModeButton: Expression | None = None
        showOptionsMenu: Expression | None = None
        showPinButton: Expression | None = None
        showSeeDataLayoutToggleButton: Expression | None = None
        showSmartNarrativeButton: Expression | None = None
        showTooltipButton: Expression | None = None
        showVisualErrorButton: Expression | None = None
        showVisualInformationButton: Expression | None = None
        showVisualWarningButton: Expression | None = None
        transparency: Expression | None = None

    properties: _VisualHeaderPropertiesHelper = Field(default_factory=_VisualHeaderPropertiesHelper)


class VisualHeaderTooltipProperties(LayoutNode):
    class _VisualHeaderTooltipPropertiesHelper(LayoutNode):
        background: Expression | None = None
        fontFamily: Expression | None = None
        fontSize: Expression | None = None
        section: Expression | None = None
        text: Expression | None = None
        themedBackground: Expression | None = None
        themedTitleFontColor: Expression | None = None
        titleFontColor: Expression | None = None
        transparency: Expression | None = None
        type: Expression | None = None
        underline: Expression | None = None

    properties: _VisualHeaderTooltipPropertiesHelper = Field(default_factory=_VisualHeaderTooltipPropertiesHelper)


class VisualLinkProperties(LayoutNode):
    class _VisualLinkPropertiesHelper(LayoutNode):
        bookmark: Expression | None = None
        disabledTooltip: Expression | None = None
        drillthroughSection: Expression | None = None
        enabledTooltip: Expression | None = None
        navigationSection: Expression | None = None
        show: Expression | None = None
        tooltip: Expression | None = None
        type: Expression | None = None
        webUrl: Expression | None = None

    properties: _VisualLinkPropertiesHelper = Field(default_factory=_VisualLinkPropertiesHelper)


class VisualTooltipProperties(LayoutNode):
    class _VisualTooltipPropertiesHelper(LayoutNode):
        background: Expression | None = None
        fontFamily: Expression | None = None
        fontSize: Expression | None = None
        section: Expression | None = None
        show: Expression | None = None
        titleFontColor: Expression | None = None
        type: Expression | None = None
        valueFontColor: Expression | None = None

    properties: _VisualTooltipPropertiesHelper = Field(default_factory=_VisualTooltipPropertiesHelper)


class VCProperties(LayoutNode):
    background: list[BackgroundProperties] | None = Field(default_factory=lambda: [BackgroundProperties()])
    border: list[BorderProperties] | None = Field(default_factory=lambda: [BorderProperties()])
    divider: list[DividerProperties] | None = Field(default_factory=lambda: [DividerProperties()])
    dropShadow: list[DropShadowProperties] | None = Field(default_factory=lambda: [DropShadowProperties()])
    general: list[GeneralProperties] | None = Field(default_factory=lambda: [GeneralProperties()])
    lockAspect: list[LockAspectProperties] | None = Field(default_factory=lambda: [LockAspectProperties()])
    spacing: list[SpacingProperties] | None = Field(default_factory=lambda: [SpacingProperties()])
    stylePreset: list[StylePresetProperties] | None = Field(default_factory=lambda: [StylePresetProperties()])
    subTitle: list[SubTitleProperties] | None = Field(default_factory=lambda: [SubTitleProperties()])
    title: list[TitleProperties] | None = Field(default_factory=lambda: [TitleProperties()])
    visualHeader: list[VisualHeaderProperties] | None = Field(default_factory=lambda: [VisualHeaderProperties()])
    visualHeaderTooltip: list[VisualHeaderTooltipProperties] | None = Field(
        default_factory=lambda: [VisualHeaderTooltipProperties()],
    )
    visualLink: list[VisualLinkProperties] | None = Field(default_factory=lambda: [VisualLinkProperties()])
    visualTooltip: list[VisualTooltipProperties] | None = Field(default_factory=lambda: [VisualTooltipProperties()])



from datetime import UTC, datetime

from pbi_core.static_files.layout._base_node import LayoutNode

PrimitiveValue = int | str | datetime | bool | None
DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S"


def parse_literal(literal_val: str) -> PrimitiveValue:
    if literal_val == "null":
        return None

    if literal_val.endswith("L"):
        return int(literal_val[:-1])

    if literal_val in {"true", "false"}:
        return literal_val == "true"

    if literal_val.startswith("datetime"):
        return datetime.strptime(literal_val[9:-1], DATETIME_FORMAT).replace(tzinfo=UTC)

    return literal_val[1:-1]


def serialize_literal(value: PrimitiveValue) -> str:
    if value is None:
        return "null"

    # int check needs to be before bool since bool is a subclass of int in Python
    if isinstance(value, int):
        return f"{value}L"

    if value in {True, False}:
        return "true" if value else "false"

    if isinstance(value, datetime):
        return f"datetime'{value.strftime(DATETIME_FORMAT)}'"

    return f"'{value}'"


class _LiteralSourceHelper(LayoutNode):
    Value: str


class LiteralSource(LayoutNode):
    Literal: _LiteralSourceHelper

    def value(self) -> PrimitiveValue:
        return parse_literal(self.Literal.Value)

    def __repr__(self) -> str:
        return f"LiteralSource({self.Literal.Value})"

    @staticmethod
    def new(value: PrimitiveValue) -> "LiteralSource":
        return LiteralSource(Literal=_LiteralSourceHelper(Value=serialize_literal(value)))



from pathlib import Path

import pydantic
from pbi_mermaid import Link, MermaidDiagram, Node

from pbi_core.static_files.layout import Layout

BASE_PATH = Path(__file__).parents[1] / "docs" / "layout" / "erd"
BASE_PATH.mkdir(parents=True, exist_ok=True)


def unwrap(t: type) -> list[type]:
    if hasattr(t, "__args__"):
        return [t2_ for t_ in t.__args__ for t2_ in unwrap(t_)]
    return [t]


BREAK_NODE_CLASSES = [
    "MeasureSource",
    "LiteralSource",
    "DateSpan",  # after LiteralSource
    "LiteralExpression",
    "HierarchyLevelSource",
    "ColumnSource",
    "AggregationSource",
    "ArithmeticSource",
    "GroupSource",
    "ScopedEvalAgg",
    "ScopedEvalArith",
    "AndCondition",
    "ComparisonCondition",
    "ContainsCondition",
    "ExistsCondition",
    "NotCondition",
    "InCondition",
    "OrCondition",
    "Condition",
    "ConditionalSource",
    "FillRule",
    "SolidExpression",
    "AlgorithmExpression",
    "AggregationExpression",
    "ColumnExpression",
    "ResourcePackageAccess",
    "ImageExpression",
    "SolidColorExpression",
    "MeasureExpression",
    "GeoJsonExpression",
    "LinearGradient2Expression",
    "LinearGradient3Expression",
    "ImageKindExpression",
    "SelectRefExpression",
    "Scope",
    "TransformMeta",
    "PrototypeQuery",
    "FilterProperties",
    "Filter",
    "Selector",
    "VCProperties",
    "ActionButton",
    "BarChart",
    "BasicShape",
    "Card",
    "ClusteredColumnChart",
    "ColumnChart",
    "ColumnProperty",
    "DonutChart",
    "Funnel",
    "Image",
    "LineChart",
    "LineStackedColumnComboChart",
    "PieChart",
    "ScatterChart",
    "Slicer",
    "TableChart",
    "Paragraph",
    "TextBox",
    "PropertyDef",
    "Query",
    "VisualFilter",
    "VisualConfig",
    "ExpansionState",
    "QueryMetadata",
    "VisualContainer",
    "ExplorationStateProperties",
    "BookmarkFilter",
    "BookmarkVisual",
    "Bookmark",
    "LayoutConfig",
    "ResourcePackage",
    "GlobalFilter",
    "Section",
    "Pod",
]


class ERD:
    visited_nodes: set[Node] = set()

    def helper(self, m: type[pydantic.BaseModel]) -> tuple[Node, set[Node], set[Link]]:
        if m.__name__ in BREAK_NODE_CLASSES:
            content = f"<a href='/layout/erd/{m.__name__}'>{m.__name__}</a>"
        else:
            content = m.__name__
        head = Node(id=m.__name__, content=content)
        nodes = {head}
        links = set()

        if head in self.visited_nodes:
            return head, nodes, links
        self.visited_nodes.add(head)

        for attr_name, attr_type in m.__pydantic_fields__.items():
            if attr_type.annotation is None:
                continue
            for child in unwrap(attr_type.annotation):
                if issubclass(child, pydantic.BaseModel):
                    sub_head, sub_nodes, sub_links = self.helper(child)
                    nodes.update(sub_nodes)
                    links.update(sub_links)
                    links.add(Link(head, sub_head, link_text=attr_name))

        return head, nodes, links

    @staticmethod
    def _get_node(nodes: set["Node"], name: str) -> "Node":
        ret = [n for n in nodes if n.id == name]
        if len(ret) != 1:
            msg = f"Node with id '{name}' not found in nodes. Found {len(ret)} matches."
            raise ValueError(msg)
        return ret[0]

    def process(self) -> None:
        self.visited_nodes.clear()
        _head, nodes, links = self.helper(Layout)
        diagram = MermaidDiagram(list(nodes), list(links), title="ERD for Layout")

        break_nodes = [self._get_node(nodes, name) for name in BREAK_NODE_CLASSES]

        diagrams = diagram.chunk(break_nodes)
        for ret_diagram in diagrams:
            with (BASE_PATH / f"{ret_diagram.title}.md").open("w", encoding="utf-8") as f:
                f.write("```mermaid\n")
                f.write(ret_diagram.to_markdown())
                f.write("\n```")


if __name__ == "__main__":
    ERD().process()
